# coding: utf-8

import os
import heapq


def top_huge_dirs_from_ncdu(topN, nodes, parent_path='', max_depth=3, depth=0):
    if isinstance(nodes, dict):
        size = nodes.get('dsize', 0)
        path = os.path.join(parent_path, nodes.get('name'))
        if depth > max_depth:
            return size
        return {path: size}

    node = nodes[0]
    subnodes = nodes[1:]
    path = os.path.join(parent_path, node.get('name'))

    if depth >= max_depth:
        ret = node.get('dsize', 0)
    else:
        ret = {}

    for subnode in subnodes:
        value = top_huge_dirs_from_ncdu(topN, subnode, path, max_depth, depth=depth + 1)
        if depth >= max_depth:
            ret += value
        else:
            ret.update(value)
            hugest_paths = heapq.nlargest(topN, ret, key=ret.get)
            ret = {p: ret.get(p) for p in hugest_paths}

    if depth == max_depth:
        return {path: ret}
    return ret


# not used now
def top_huge_files_from_ncdu(path='', blocks=[], res={}):
    dir_name = blocks[0].get('name')
    if dir_name.startswith('/'):
        path += dir_name
    else:
        path += '/' + dir_name

    for idx, block in enumerate(blocks):
        if isinstance(block, list):
            top_huge_files_from_ncdu(path, block, res)

        if isinstance(block, dict):
            if idx == 0:
                file_name = path
            else:
                file_name = path + '/' + block.get('name')
            # Directories in xfs do not have 'dsize' key. (?)
            size = block.get('dsize', 0)
            res[file_name] = size

    return res



# coding: utf-8

import json
import urllib
import logging

from sa_tools_core.utils import get_config

PUSHOVER_URL = 'https://api.pushover.net/1/messages.json'

logger = logging.getLogger(__name__)


def send_message(user_key, message):
    # APP_TOKEN created on pushover.net
    APP_TOKEN = get_config('pushover')

    data = {
        "token": APP_TOKEN,
        "user": user_key,
        "message": message,
    }
    payload = urllib.urlencode(data)

    resp = json.loads(urllib.urlopen(PUSHOVER_URL, payload).read())
    if int(resp['status']) != 1:
        raise Exception('api failed, resp: %s' % resp)



#!/usr/bin/env python
# coding: utf-8

from mako.lookup import TemplateLookup

from sa_tools_core.consts import PROJECT_ROOT
from sa_tools_core.utils import to_unicode


TEMPLATES_DIR = PROJECT_ROOT + "/templates"
MAGIC_EMPTY_LINE_MARK = "|"
lookup = TemplateLookup(
    directories=[TEMPLATES_DIR],
    input_encoding="utf8",
    output_encoding="utf8",
    encoding_errors="ignore",
)


def render(template_name, strip_empty_lines=True, **data):
    out = lookup.get_template(template_name).render(**data)
    # When using Python 3, the render() method will return a bytes object,
    # if output_encoding is set. Otherwise it returns a string.
    out = to_unicode(out)

    if strip_empty_lines:
        return (
            "\n".join(
                [
                    ("" if i.strip() == MAGIC_EMPTY_LINE_MARK else i)
                    for i in out.splitlines()
                    if i.strip()
                ]
            )
            + "\n"
        )
    else:
        return out


def render_notification(**kw):
    title = render("icinga/notification.mako", fragment_type="title", **kw)
    content = render("icinga/notification.mako", fragment_type="content", **kw)
    return title, content



# coding: utf-8

import re
import sys
import logging
import argparse

from sa_tools_core.libs.sentry import report
from sa_tools_core.libs.permission import require_user
from sa_tools_core.libs.notify.email import send_mail
from sa_tools_core.libs.notify.wework import send_message as send_wework
from sa_tools_core.libs.notify.lark import send_message as send_lark
from sa_tools_core.libs.notify.pushbullet import send_message as send_pushbullet
from sa_tools_core.libs.notify.pushover import send_message as send_pushover
from sa_tools_core.libs.notify.telegram import send_message as send_telegram
from sa_tools_core.libs.notify.sms import send_sms
from sa_tools_core.utils import get_os_username
from sa_tools_core.consts import SYSADMIN_EMAIL

logger = logging.getLogger(__name__)
logging.getLogger("requests").setLevel(logging.WARN)

NOTIFY_TYPES = ("wework", "lark", "email", "sms", "pushbullet", "pushover", "telegram")
DEFAULT_TITLE = "Sent from sa-notify"
DEFAULT_MSG_TYPE = "text"


class Notifier(object):
    def __init__(
        self,
        content=None,
        title=DEFAULT_TITLE,
        from_addr=SYSADMIN_EMAIL,
        msg_type=DEFAULT_MSG_TYPE,
    ):
        self.content = content
        self.title = title
        self.from_addr = from_addr
        self.msg_type = msg_type

    def __getattr__(self, attr):
        if attr in NOTIFY_TYPES:
            f = getattr(self, "_%s" % attr)

            def _func(*a, **kw):
                addrs = kw.get("addrs") or a[0]
                addrs = [addr for addr in addrs if addr]
                if not addrs:
                    logger.warning("notify abort, ignoring empty %s addrs", attr)
                    return
                kw["content"] = kw.get("content") or self.content
                kw["title"] = kw.get("title") or self.title
                if kw["title"]:
                    kw["title"] = kw["title"].strip()
                logger.info("notify %s %s", attr, addrs)
                return f(*a, **kw)

            return _func
        raise AttributeError

    def _wework(self, addrs, content=None, **kw):
        content = content if content is not None else self.content
        send_wework(addrs, content, msg_type=self.msg_type)

    def _lark(self, addrs, content=None, **kw):
        send_lark(addrs, content, **kw)

    def _email(self, addrs, content=None, title=None, from_addr=None, **kw):
        content = content if content is not None else self.content
        send_mail(addrs, content, subject=title, from_addr=from_addr or self.from_addr)

    def _pushbullet(self, addrs, content=None, title=None, **kw):
        content = content if content is not None else self.content
        for addr in addrs:
            send_pushbullet(addr, title, content)

    def _pushover(self, addrs, content=None, **kw):
        content = content if content is not None else self.content
        for addr in addrs:
            send_pushover(addr, content)

    def _telegram(self, addrs, content=None, **kw):
        content = content if content is not None else self.content
        for addr in addrs:
            send_telegram(addr, content, msg_type=self.msg_type)

    def _sms(self, addrs, content=None, **kw):
        content = content if content is not None else self.content
        for addr in addrs:
            send_sms(addr, content)


@require_user
def notify(args):
    msg_type = "markdown" if args.markdown else "text"
    notifier = Notifier(args.content, args.subject, msg_type=msg_type)
    for type_ in NOTIFY_TYPES:
        values = vars(args)[type_]
        if values:
            addrs = [i for v in values for i in re.split(r"[,\s]+", v)]
            try:
                getattr(notifier, type_)(addrs, from_addr=args.from_addr, company=args.company)
            except Exception as e:
                logger.exception("Notifier.%s(%s) failed: %s", type_, addrs, e)
                report()


def main(args=None):
    """
    e.g.
    $ sa-notify --wework user1 --content 'xxx'
    $ echo 'xxx' | sa-notify --wework user1,user2 --email user1@douban.com user3@douban.com
    """
    parser = argparse.ArgumentParser(
        epilog=main.__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("-u", "--user", help="LDAP username, use your OS user name by default.")
    parser.add_argument("-c", "--content", help="content.")
    parser.add_argument(
        "-s", "--subject", help='subject. default="%(default)s"', default=DEFAULT_TITLE
    )
    parser.add_argument("-f", "--from-addr", help="From address, currently only works for email.")
    parser.add_argument("--company", help="Company user in, used when multiple company or tenant is configured.")
    for type_ in NOTIFY_TYPES:
        parser.add_argument(
            "--%s" % type_, nargs="*", help="your enterprise address of %s." % type_
        )
    parser.add_argument(
        "--markdown",
        help="use markdown rendering, only wework & telegram supported",
        action="store_true",
    )

    args = parser.parse_args(args)
    args.user = args.user or get_os_username()

    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(name)s %(levelname)s %(message)s")

    if not any(vars(args)[type_] for type_ in NOTIFY_TYPES):
        parser.error("too few arguments")
    if args.content is None:
        args.content = sys.stdin.read()
        if not args.content:
            parser.error("content empty")

    notify(args)



# coding: utf-8

import logging
import requests

from sa_tools_core.utils import get_config
from sa_tools_core.consts import PROXIES

logger = logging.getLogger(__name__)


def send_message(chatid, content, msg_type="text"):
    PUSH_KEY = get_config("telegram")
    PUSH_URL = "https://api.telegram.org/bot%s/sendMessage" % PUSH_KEY

    message = {
        "chat_id": chatid,
        "text": content,
    }
    if msg_type == "markdown":
        message["parse_mode"] = "Markdown"
    elif msg_type == "html":
        message["parse_mode"] = "HTML"

    req = requests.post(PUSH_URL, json=message, proxies=PROXIES)
    resp = req.json()
    if not resp["ok"]:
        raise Exception("api failed, resp: %s" % resp)



# coding: utf-8

from __future__ import print_function

import re
import cmd
import sys
import shlex
import string
import logging
import subprocess
from fnmatch import fnmatch
from argparse import ArgumentParser, FileType
from subprocess import Popen, PIPE
from multiprocessing import Value
from textwrap import TextWrapper
from operator import itemgetter

from progressbar import ProgressBar, Percentage, Timer, Bar
from terminaltables import AsciiTable

from sa_tools_core.libs.editor import Editor
from sa_tools_core.utils import reverse_func

from sa_tools_core.libs.ansible import DefaultRunnerCallbacks, Inventory, Runner

logger = logging.getLogger(__name__)

DEFAULT_FORKS = 20
RESULT_FIELDS = ('host', 'rc', 'stdout', 'stderr')
FIELDS_ORDER = ('host', 'stdout', 'stderr', 'rc')


def get_wrap():
    t_width = int(subprocess.check_output(['tput', 'cols']).strip())
    wrap_width = int(t_width / 3)
    return TextWrapper(width=wrap_width).wrap


class ScriptCLI(cmd.Cmd, object):
    PROMPT_TMPL = '(sa-script) {filters} $ '

    parser = ArgumentParser()

    parser.add_argument('host_pattern', metavar='<host-pattern>', nargs='?', help='host pattern that ansible takes')
    parser.add_argument('-f', '--file', type=FileType('r'), dest='script', default=sys.stdin, help='script file name')
    parser.add_argument('-u', '--user', metavar='REMOTE_USER', help='connect as this user')
    parser.add_argument(
        '-c', '--connection', metavar='CONNECTION', default='ssh', help='connection type to use (default=%(default)s)')
    parser.add_argument('-s', '--sudo', action='store_true', help='run operations with sudo')
    parser.add_argument('-F', '--forks', metavar='NUM', type=int, default=DEFAULT_FORKS, help='Level of parallelism')
    parser.add_argument(
        '-C',
        '--compatible',
        action='store_true',
        help='run in compatible mode, use ansible.runner instead of doa.runner as runner class')
    parser.add_argument('-r', '--retry', metavar='NUM', type=int, default=0, help='Retry NUM times(dark only)')
    parser.add_argument('-R', '--retry-failed', action='store_true', help='Also retry contacted_failed')
    parser.add_argument('--host', help='filter results match the host pattern')
    parser.add_argument('--rc', help='filter results match the rc pattern')
    parser.add_argument('--stdout', help='filter results match the stdout pattern')
    parser.add_argument('--stderr', help='filter results match the stderr pattern')
    parser.add_argument('-q', '--quit', action='store_true', help='quit after script execute')

    def __init__(self):
        super(ScriptCLI, self).__init__()
        self.filter_args = []
        self.mark = {}

        self.parser.usage = """
sa-script [-h] [-s] [-f NUM] [-r NUM] [-R] [--host HOST] [--rc RC]
                 [--stdout STDOUT] [--stderr STDERR] [-q]
                 <host-pattern> [script]"""

        args = self.parser.parse_args()
        if not args.host_pattern:
            self.parser.error('too few arguments')

        self.parser.usage = None

        try:
            self._run(args)
        except KeyboardInterrupt:
            sys.exit(2)
        except (ScriptEmptyError, NoHostsError) as e:
            print(e, file=sys.stderr)
            sys.exit(2)

        if args.script == sys.stdin and not args.script.isatty():
            args.quit = True

        if not args.quit:
            while True:
                try:
                    self.cmdloop()
                except KeyboardInterrupt:
                    self.onecmd('reset')
                    print()

    def postloop(self):
        sys.exit(0)

    def _run(self, args):
        self.runner = ScriptRunner(
            args.host_pattern,
            args.script,
            args.user,
            args.connection,
            args.sudo,
            args.retry,
            args.retry_failed,
            args.forks,
            args.compatible,
        )

        self.runner.run()

        if args.host:
            self.onecmd('host ' + args.host)
        if args.rc:
            self.onecmd('rc ' + args.rc)
        if args.stdout:
            self.onecmd('stdout ' + args.stdout)
        if args.host:
            self.onecmd('stderr ' + args.stderr)

        self.onecmd('reset')  # init prompt and mark
        self.onecmd('print')

    def do_run(self, line):
        try:
            args = self.parser.parse_args(shlex.split(line))
        except SystemExit:
            pass  # ArgumentParser will print error message
        else:
            args.host_pattern = args.host_pattern or ':'.join(self.mark)
            try:
                self._run(args)
            except KeyboardInterrupt:
                print('\nabort...')
            except (ScriptEmptyError, NoHostsError) as e:
                print(e, file=sys.stderr)

    def help_run(self):
        print(
            """
run new script and print result
================================
(sa-script) $ run [host-pattern] [script filename] [args]

    empty host-pattern will use last host-pattern

e.g.:
    (sa-script) $ run
    (sa-script) $ run all
    (sa-script) $ run all xxx.sh
              """)

    def do_host(self, line):
        self.marker('host', line)

    def help_host(self):
        print(self.get_filter_help('host'))

    def do_rc(self, line):
        self.marker('rc', line)

    def help_rc(self):
        print(self.get_filter_help('rc'))

    def do_stdout(self, line):
        self.marker('stdout', line)

    def help_stdout(self):
        print(self.get_filter_help('stdout'))

    def do_stderr(self, line):
        self.marker('stderr', line)

    def help_stderr(self):
        print(self.get_filter_help('stderr'))

    def do_reset(self, line):
        self.filter_args = []
        self.reset_mark()

    def help_reset(self):
        print('reset filter')

    def _output(self, data, cmd=None):
        if cmd:
            pipe = Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = pipe.communicate(input=data)

            if out:
                print(out)
            if err:
                print(err, file=sys.stderr)
        else:
            print(data)

    def do_print(self, line):
        if '|' in line:
            fields, cmd = line.split('|', 1)
        else:
            fields = line
            cmd = None

        fields = fields.split()
        fields = fields or ['host', 'rc', 'stdout', 'stderr']

        if not all(field in RESULT_FIELDS for field in fields):
            print('no such fields, check your input', file=sys.stderr)
            return

        data = self.list_fields(*fields)

        if len(fields) > 1:
            header = [fields]
            table = AsciiTable(header + data)

            if not table.ok:
                # the table not fits within the terminal width
                wrap = get_wrap()
                data = [['\n'.join(wrap(i)) for i in item] for item in data]
                table = AsciiTable(header + data)

            data = table.table
        else:
            data = '\n'.join([item[0] for item in data])
            if not data.endswith('\n'):
                data += '\n'

        self._output(data, cmd)

    do_p = do_print

    def help_print(self):
        print(
            """
print marked results (and pipe to shell commands)
==================================================
(sa-script) $ print [filed] [| shell commands]

e.g.:
    (sa-script) $ print
    (sa-script) $ print stderr
    (sa-script) $ print | cat > dump
    (sa-script) $ print stdout | grep xxx | sed xxx | cat > dump
              """)

    help_p = help_print

    def do_exit(self, line):
        return True

    def help_exit(self):
        print('exit')

    def do_EOF(self, line):
        print()
        return True

    help_EOF = help_exit

    def emptyline(self):
        pass

    def reset_mark(self):
        self.mark = dict.fromkeys(self.runner.results.keys(), True)

    def marker(self, key, pattern):
        reverse = False
        if pattern.startswith('!'):
            reverse = True
            pattern = pattern[1:]

        if pattern.startswith(('~', '=')):
            op = pattern[0]
            pattern = pattern[1:]
        elif pattern:
            op = '~'
        else:
            op = ''

        pattern = pattern.strip()

        if op == '=':
            pred_func = lambda result: fnmatch(result.get(key), pattern)  # NOQA
        elif op == '~':
            pred_func = lambda result: re.search(pattern, result.get(key))  # NOQA
        else:
            # pattern is None or empty string
            pred_func = lambda result: result.get(key)  # NOQA

        if reverse:
            pred_func = reverse_func(pred_func)

        self.filter_args.append(
            '{key}{r}{op}{pattern}'.format(key=key, r='!' if reverse else '', op=op, pattern=pattern))

        try:
            self.mark = {
                host: True for host, result in self.runner.results.items() if self.mark.get(host) and pred_func(result)
            }
        except re.error as e:
            self.onecmd('reset')
            print(e, file=sys.stderr)

    def list_fields(self, *fields):
        marked_results = [self.runner.results.get(k) for k, v in self.mark.items() if v]

        rtn = [[result.get(field, '') for field in fields] for result in marked_results]

        for field in FIELDS_ORDER:
            if field in fields:
                rtn.sort(key=itemgetter(fields.index(field)))

        return rtn

    def get_filter_help(self, field):
        return """
filter {field} field
======================================
(sa-script) $ {field} [pattern]
    =<pattern> for wildcard filter.
    [~]<pattern> for regex filter.
    empty pattern for non empty filter""".format(field=field)

    @property
    def prompt(self):
        return self.PROMPT_TMPL.format(filters=' '.join(self.filter_args))


class ScriptRunnerCallbacks(DefaultRunnerCallbacks):
    def __init__(self, pbar):
        self.pbar = pbar
        self.counter = Value('i', 0)
        super(ScriptRunnerCallbacks, self).__init__()

    def on_failed(self, host, res, ignore_errors=False):
        self.update_pbar()

    def on_ok(self, host, res):
        self.update_pbar()

    def on_skipped(self, host, item=None):
        logger.warning('{host} skipped'.format(host=host))

    def on_unreachable(self, host, res):
        self.update_pbar()

    def on_no_hosts(self):
        print('no hosts matched\n', file=sys.stderr)

    def update_pbar(self):
        with self.counter.get_lock():
            self.counter.value += 1
            self.pbar.update(self.counter.value)


class ScriptEmptyError(Exception):
    def __str__(self):
        return 'the script you provide is empty'


class NoHostsError(Exception):
    def __str__(self):
        return 'no hosts matched'


class ScriptRunner(object):
    inventory = Inventory()
    editor = Editor(extension='.sh')

    def __init__(
        self,
        host_pattern,
        script=sys.stdin,
        user=None,
        connection='ssh',
        sudo=False,
        retry=0,
        retry_failed=False,
        forks=DEFAULT_FORKS,
        compatible=False,
    ):

        self.host_pattern = host_pattern

        if script == sys.stdin and script.isatty():
            self.script = self.editor.edit()
            print(self.script)
        else:
            self.script = script.read()

        self.script = ''.join([c for c in self.script if c in string.printable]) if self.script else ''

        if not self.script:
            raise ScriptEmptyError

        self.user = user
        self.connection = connection
        self.sudo = sudo
        self.max_retries = retry
        self.retry_failed = retry_failed
        self.forks = forks
        self.compatible = compatible

        self._results = None
        self.results_is_processed = False

    def _run(self, host_pattern):
        run_hosts = self.inventory.list_hosts(host_pattern)

        if not run_hosts:
            raise NoHostsError

        widgets = [Percentage(), Bar(), Timer()]
        pbar = ProgressBar(widgets=widgets, max_value=len(run_hosts))

        pbar.start()
        params = dict(
            module_name='shell',
            module_args=self.script,
            remote_user=self.user,
            connection=self.connection,
            become=self.sudo,
            callbacks=ScriptRunnerCallbacks(pbar),
            run_hosts=host_pattern,
            forks=self.forks)
        results = Runner(**params).run()
        pbar.finish()
        print()

        _results = results.get('contacted', {})
        _results.update(results.get('dark', {}))
        return _results

    def run(self):
        print('Executing...')

        self._results = self._run(self.host_pattern)

        self.retry()

    def retry(self):

        pred_failed = lambda result: result.get('rc', None) != 0  # NOQA
        pred_dark = lambda result: result.get('rc', None) is None  # NOQA

        pred_func = pred_failed if self.retry_failed else pred_dark

        for i in range(self.max_retries):
            need_retry = [host for host, result in self._results.items() if pred_func(result)]

            print('Retrying {}...'.format(i + 1))

            results = self._run(':'.join(need_retry))

            _results = results.get('contacted', {})
            _results.update(results.get('dark', {}))
            self._results.update(_results)

    @property
    def results(self):
        if not self.results_is_processed:
            for host, result in self._results.items():
                result.update({'host': host})
                result['rc'] = str(result.get('rc', ''))
                result['stdout'] = result.get('stdout', '')
                result['stderr'] = result.get('stderr', result.get('msg', '') + result.get('exception', ''))

            self.results_is_processed = True

        return self._results


def main():
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(logging.Formatter('%(levelname)s - %(message)s'))
    logger.addHandler(ch)

    ScriptCLI()



# coding: utf-8

import logging
import requests

from sa_tools_core.consts import NOTIFICATION_GATEWAY_API, NOTIFICATION_GATEWAY_TIMEOUT

logger = logging.getLogger(__name__)


def add_notification(type_, host, hoststate, service,
                     content, method, addrs, duration):
    payload = dict(type=type_,
                   host=host, hoststate=hoststate,
                   service=service,
                   content=content,
                   method=method, addrs=addrs, duration=duration)
    resp = requests.post(NOTIFICATION_GATEWAY_API,
                         data=payload,
                         timeout=NOTIFICATION_GATEWAY_TIMEOUT)
    resp.raise_for_status()
    resp_json = resp.json()
    logger.info(resp_json)
    return resp_json['ok']



# coding: utf-8

from __future__ import print_function

import json
import logging
import argparse
from six.moves.configparser import ConfigParser

from pydnspod import pydnspod

from sa_tools_core.libs.permission import require_sa
from sa_tools_core.consts import (EXTERNAL_DOMAINS_CONFIG_FILE,
                                  DNS_MONITOR_CALLBACK_URL,
                                  DEFAULT_DNS_DOMAIN)
from sa_tools_core.utils import (get_os_username, get_config,
                                 ipv6_addr_to_tinydns_generic,
                                 output, to_str)

logger = logging.getLogger(__name__)

DOMAIN = DEFAULT_DNS_DOMAIN
DEFAULT_TTL = 300
DEFAULT_MX = 10
DEFAULT_LINE = '默认'
DNSPOD_TO_TINYDNS = {
    'A': ['={domain}:{value}:{ttl}', '+{domain}:{value}:{ttl}'],
    'AAAA': ':{domain}:28:{value}:{ttl}',
    'MX': '@{domain}::{value}:{dist}:{ttl}',
    'NS': '.{domain}::{value}:{ttl}',
    'TXT': '\'{domain}:{value}:{ttl}',
    'CNAME': 'C{domain}:{value}:{ttl}',
}
MAX_RECORDS_COUNT = 3000

DNSPOD_TIMEOUT = 5
DNSPOD_RETRIES = 3

dns_monitor_cb_key_func = lambda: get_config('dns_monitor_callback_key')  # NOQA

external_domains_config = ConfigParser(allow_no_value=True)
external_domains_config.read(EXTERNAL_DOMAINS_CONFIG_FILE)


def get_api_token_by_domain(domain):
    """
    config file example:
    token_id,token,domain
    111,yyy,*
    222,xxx,domain1.com|domain2.com|domain3.co
    """
    result = {}
    for conf in get_config('dnspod').splitlines():
        id_token, domains = conf.rsplit(',', 1)
        for scope in domains.split('|'):
            result[scope] = id_token
    logger.debug("domain config is %s", result)
    return result[domain] if domain in result else result['*']


class DNSPod(object):
    """manipulate the domain served by dnspod"""
    def __init__(self, login_email="", login_password="",
                 api_token_or_func=get_api_token_by_domain,
                 domain=DOMAIN,
                 dry_run=False,
                 verbose=True):
        self.email = login_email
        self.passwd = login_password
        self.api_token = api_token_or_func(domain) if callable(api_token_or_func) else api_token_or_func
        self.dry_run = dry_run
        self.verbose = verbose
        self._dns_monitor_cb_key = None

        self.domain_str = domain
        domain_dict = json.loads(self.domain.info(domain=domain))['domain']
        self.domain_id = domain_dict['id']
        self.domain_grade = domain_dict['grade']
        self.domain_grade_title = domain_dict['grade_title']

    @property
    def user(self):
        return pydnspod.User(self.email, self.passwd, self.api_token,
                             timeout=DNSPOD_TIMEOUT,
                             retries=DNSPOD_RETRIES)

    @property
    def domain(self):
        """
        dnspod 返回码说明 (http://www.dnspod.cn/docs/info.html#common-response)
        :todo 这块放入pydnspod模块做详细处理
        -1 登陆失败
        -2 API使用超出限制
        -3 不是合法代理 (仅用于代理接口)
        -4 不在代理名下 (仅用于代理接口)
        -7 无权使用此接口
        -8 登录失败次数过多，帐号被暂时封禁
        85 帐号异地登录，请求被拒绝
        -99 此功能暂停开放，请稍候重试
        1 操作成功
        2 只允许POST方法
        3 未知错误
        6 用户ID错误 (仅用于代理接口)
        7 用户不在您名下 (仅用于代理接口)
        83 该帐户已经被锁定，无法进行任何操作
        85 该帐户开启了登录区域保护，当前IP不在允许的区域内
        """
        return pydnspod.Domain(self.email, self.passwd, self.api_token,
                               timeout=DNSPOD_TIMEOUT,
                               retries=DNSPOD_RETRIES)

    @property
    def record(self):
        return pydnspod.Record(self.email, self.passwd, self.api_token,
                               timeout=DNSPOD_TIMEOUT,
                               retries=DNSPOD_RETRIES)

    @property
    def monitor(self):
        return pydnspod.Monitor(self.email, self.passwd, self.api_token,
                                timeout=DNSPOD_TIMEOUT,
                                retries=DNSPOD_RETRIES)

    @property
    def dns_monitor_cb_key(self):
        if self._dns_monitor_cb_key is None:
            self._dns_monitor_cb_key = dns_monitor_cb_key_func()
        return self._dns_monitor_cb_key

    def get_user_log(self):
        ret = self.user.log()
        return json.loads(ret).get('log', [])

    def get_record_types(self):
        ret = self.domain.record_type(domain_grade=self.domain_grade)
        return json.loads(ret).get('types', [])

    def get_record_lines(self):
        ret = self.domain.record_line(domain_id=self.domain_id,
                                      domain_grade=self.domain_grade)
        return json.loads(ret).get('lines', [])

    def get_records(self, **kw):
        # FIXME: https://www.dnspod.cn/docs/records.html#record-list, iter to get all
        records = json.loads(self.record.list(domain_id=self.domain_id, length=MAX_RECORDS_COUNT, **kw)).get('records', [])
        return records

    def remove_records(self, sub_domain, type=None, value=None, line=DEFAULT_LINE):
        """remove all records of the sub_domain, type, value"""
        line = line if line else DEFAULT_LINE
        records = self.get_records(sub_domain=sub_domain)
        # NOTE: only deal with default line type records by default
        exist_records = [r for r in records
                         if (to_str(r['line']) == line)
                         and (not type or r['type'] == type)
                         and (not value or r['value'].rstrip('.') == value.rstrip('.'))]

        if self.verbose:
            logger.info('remove_records({sub_domain}, {type}, {value}, {line})'
                        .format(sub_domain=sub_domain,
                                type=type, value=value, line=line))

        if exist_records:
            if self.verbose:
                [logger.info('self.record.remove({domain_id}, {record_id})'
                             .format(domain_id=self.domain_id, record_id=r['id']))
                 for r in exist_records]
            if self.dry_run:
                return [], ''

            results = [json.loads(self.record.remove(domain_id=self.domain_id, record_id=r['id']))
                       for r in exist_records]

            successes = [int(r['status']['code']) == 1 for r in results]
            msg = 'remove records failed: {msgs}'.format(msgs=' || '.join([r['status']['message']
                                                         for r in results]))
        else:
            return [], ''

        return exist_records, '' if all(successes) else msg

    def add_or_modify_record(self, sub_domain, type, value, ttl=DEFAULT_TTL, mx=DEFAULT_MX,
                             line=DEFAULT_LINE,
                             force_enable=None):
        """
        - add or modify 一个 记录
            - 若存在，则 modify
                - 若没差，则直接返回
                - 若有差，
                    - 若 enable，则先把对立的 type 记录都设 disable
                        （若为 CNAME，则把其他 CNAME 都设为 disable），再真正调 modify
                    - 若 disable，真正调 modify
            - 若不存在，则 add
                - 若 enable，则先把对立的 type 记录都设 disable
                    （若为 CNAME，则把其他 CNAME 都设为 disable），再真正调 add
                - 若 disable，真正调 add
        """
        ttl = ttl if ttl else DEFAULT_TTL
        line = line if line else DEFAULT_LINE
        mx = mx if mx else DEFAULT_MX
        mx = "0" if type != 'MX' else mx
        status = 'enable' if force_enable else 'disable'

        # @.example.com.h1.aqb.so. -> example.com.h1.aqb.so.
        value = value.replace('@', '').lstrip('.')

        records = self.get_records(sub_domain=sub_domain)
        # NOTE: only deal with default line type records by default
        exist_records = [r for r in records
                         if to_str(r['line']) == line
                         and r['type'] == type
                         and r['value'].rstrip('.') == value.rstrip('.')]

        if self.verbose:
            logger.info('add_or_modify_record({sub_domain}, {type}, {value}, ttl={ttl}, '
                        'mx={mx}, line={line}, force_enable={force_enable})'
                        .format(sub_domain=sub_domain,
                                type=type, value=value,
                                ttl=ttl, mx=mx, line=line, force_enable=force_enable))

        if exist_records:

            # if no changes
            if all([r['ttl'] == str(ttl)
                    and r['mx'] == str(mx)
                    and ('enabled' if int(r['enabled']) else 'disabled') == status
                    for r in exist_records]):
                return exist_records, ''

            if force_enable:
                if type == 'A':
                    self.set_status(sub_domain, 'CNAME', enable=False)
                elif type == 'CNAME':
                    self.set_status(sub_domain, 'A', enable=False)
                    self.set_status(sub_domain, 'CNAME', enable=False, except_values=[value])

            # if only change status
            only_status_change_exist_records = [r for r in exist_records
                                                if (r['ttl'] == str(ttl)
                                                    and r['mx'] == str(mx)
                                                    and ('enabled' if int(r['enabled'])
                                                         else 'disabled') != status)]
            rest_exist_records = [r for r in exist_records
                                  if not (r['ttl'] == str(ttl)
                                          and r['mx'] == str(mx)
                                          and ('enabled' if int(r['enabled'])
                                               else 'disabled') != status)]
            self.set_status_by_records(only_status_change_exist_records, enable=force_enable)

            if self.verbose:
                [logger.info('record.modify({domain_id}, {record_id}, {sub_domain}, '
                             '{record_type}, {value}, {record_line}, {ttl}, {mx}, {status})'
                             .format(domain_id=self.domain_id, record_id=r['id'],
                                     sub_domain=r['name'],
                                     record_type=type, value=value,
                                     record_line=line,
                                     ttl=ttl, mx=mx, status=status))
                 for r in rest_exist_records]
            if self.dry_run:
                return [], ''

            # NOTE: 一小时超过五次没有任何变动的修改会导致记录被锁定一小时，会报 API usage is limited
            results = [json.loads(self.record.modify(domain_id=self.domain_id, record_id=r['id'],
                                                     sub_domain=r['name'],
                                                     record_type=type, value=value,
                                                     record_line=line,
                                                     ttl=ttl, mx=mx, status=status))
                       for r in rest_exist_records]

            successes = [int(r['status']['code']) == 1 for r in results]
            msg = 'modify records failed: {msgs}'.format(msgs=' || '.join([r['status']['message']
                                                         for r in results]))
            return exist_records, '' if all(successes) else msg
        else:
            if force_enable:
                if type == 'A':
                    self.set_status(sub_domain, 'CNAME', enable=False)
                elif type == 'CNAME':
                    self.set_status(sub_domain, 'A', enable=False)
                    self.set_status(sub_domain, 'CNAME', enable=False)

            if self.verbose:
                logger.info('record.create({domain_id}, {sub_domain}, {record_type}, {value}, '
                            '{record_line}, {ttl}, {mx}, {status})'
                            .format(domain_id=self.domain_id, sub_domain=sub_domain,
                                    record_type=type, value=value, record_line=line,
                                    ttl=ttl, mx=mx, status=status))
            if self.dry_run:
                return [], ''

            result = json.loads(self.record.create(domain_id=self.domain_id, sub_domain=sub_domain,
                                                   record_type=type, value=value, record_line=line,
                                                   ttl=ttl, mx=mx, status=status))
            success = int(result['status']['code']) == 1

            msg = 'add record failed: {msg}'.format(msg=result['status']['message'])
            return [result['record']] if success else [], '' if success else msg

    def set_status(self, sub_domain, type, value=None, enable=True, except_values=[]):
        """set status of records which has same type (and value)"""
        except_values = [v.rstrip('.') for v in except_values]

        records = self.get_records(sub_domain=sub_domain)
        if value:
            exist_records = [r for r in records
                             if r['type'] == type and r['value'].rstrip('.') == value.rstrip('.')]
        else:
            exist_records = [r for r in records
                             if r['type'] == type and r['value'].rstrip('.') not in except_values]

        if exist_records:
            return self.set_status_by_records(exist_records, enable=enable)
        else:
            return False, 'the record not found'

    def set_status_by_records(self, records, enable=True):
        status = 'enable' if enable else 'disable'

        if self.verbose:
            [logger.info('record.modify_status({domain_id}, {record_id}, {status})'
                         '  # {sub_domain}, {record_type}, {value}'
                         .format(domain_id=self.domain_id,
                                 record_id=r['id'],
                                 status=status,
                                 sub_domain=r['name'],
                                 record_type=r.get('type', ''),
                                 value=r.get('value', '')))
             for r in records]
        if self.dry_run:
            return True, ''

        results = [json.loads(self.record.modify_status(domain_id=self.domain_id,
                                                        record_id=r['id'],
                                                        status=status))
                   for r in records]
        successes = [int(r['status']['code']) == 1 for r in results]
        msg = 'set record status failed: {msgs}'.format(msgs=' || '.join([r['status']['message']
                                                        for r in results]))
        return all(successes), '' if all(successes) else msg

    def get_monitors(self, **kwargs):
        return json.loads(self.monitor.list(**kwargs)).get('monitors', [])

    def add_monitors(self, sub_domain, port, monitor_interval, monitor_schema_type, monitor_uri_path, points,
                     bak_ip_mode, send_to_dns_monitor_cb=True):
        records = [r for r in self.get_records(sub_domain=sub_domain) if r['enabled'] == '1' and
                   (r['type'] == 'A' or r['type'] == 'CNAME')]

        if self.verbose:
            for r in records:
                logger.info('add_monitor({sub_domain} {line} {type} {value})'.format(
                            sub_domain=r['name'], line=to_str(r['line']),
                            type=r['type'], value=r['value']))

        if self.dry_run:
            return

        params = {
            'domain_id': self.domain_id,
            'port': port,
            'monitor_interval': monitor_interval,
            'host': '{}.{}'.format(sub_domain, self.domain_str),
            'monitor_type': monitor_schema_type,
            'monitor_path': monitor_uri_path,
            'points': points,
            'bak_ip': bak_ip_mode,
            'email_notice': 'me',
            'less_notice': 'yes'
        }

        if send_to_dns_monitor_cb:
            params.update({'callback_url': DNS_MONITOR_CALLBACK_URL.format(cb_token=self.dns_monitor_cb_key),
                          'callback_key': self.dns_monitor_cb_key})

        for r in records:
            params['record_id'] = r['id']
            self.monitor.create(**params)

    def remove_monitors(self, sub_domain):
        monitors = [m for m in self.get_monitors() if m['domain_id'] == self.domain_id and m['sub_domain'] == sub_domain]

        if self.verbose:
            for m in monitors:
                logger.info('remove_monitors({sub_domain} {line} {ip} {type} {path})'.format(
                            sub_domain=sub_domain, line=to_str(m['record_line']), ip=m['ip'], type=m['monitor_type'],
                            path=m['monitor_path']))

        if self.dry_run:
            return

        for m in monitors:
            self.monitor.remove(monitor_id=m['monitor_id'])


def parse_sub_domains(sub_domains, top_level_domain):
    domain_group_prefix = '' if top_level_domain == DOMAIN else top_level_domain + ':'

    sub_domains = sub_domains.split(',')
    sub_domains = [s.strip() for s in sub_domains if s.strip()]

    sub_domains_extended = [sub_domain
                            for domain_group in sub_domains
                            if external_domains_config.has_section(domain_group_prefix + domain_group)
                            for sub_domain in external_domains_config.options(domain_group_prefix + domain_group)]
    sub_domains = [sub_domain
                   for sub_domain in sub_domains
                   if not external_domains_config.has_section(domain_group_prefix + sub_domain)] + sub_domains_extended
    return sub_domains


@require_sa
def list_monitor(args):
    dnspod = DNSPod(domain=args.domain)

    monitors = dnspod.get_monitors()

    monitors = [m for m in monitors if m['domain'] == args.domain]
    if not args.all:
        monitors = [m for m in monitors if m['monitor_status'] == 'enabled']
    if args.sub_domain:
        monitors = [m for m in monitors if m['sub_domain'] == args.sub_domain]
    output('{:>10} {:<12} {:>15} {:>4} {:>5} {:<20} {:<5}'.format('sub_domain', 'line', 'ip', 'port', 'type', 'path',
                                                                  'status'))
    output('-' * 80)
    for monitor in monitors:
        output('{:>10} {:<12} {:>15} {:>4} {:>5} {:<20} {:<5}'.format(monitor['sub_domain'],
                                                                      to_str(monitor['record_line']),
                                                                      monitor['ip'], monitor['port'], monitor['monitor_type'],
                                                                      monitor['monitor_path'],
                                                                      monitor['monitor_status']))


@require_sa
def add_monitor(args):
    dnspod = DNSPod(dry_run=args.dry_run, domain=args.domain)

    sub_domains = parse_sub_domains(args.sub_domains, args.domain)

    for sub_domain in sub_domains:
        dnspod.add_monitors(sub_domain, args.port, args.monitor_interval, args.monitor_schema_type,
                            args.monitor_uri_path, args.points, args.bak_ip_mode, args.send_to_callback)


@require_sa
def remove_monitor(args):
    dnspod = DNSPod(dry_run=args.dry_run, domain=args.domain)

    sub_domains = parse_sub_domains(args.sub_domains, args.domain)

    for sub_domain in sub_domains:
        dnspod.remove_monitors(sub_domain)


@require_sa
def list_(args):
    dnspod = DNSPod(domain=args.domain)
    params = {}
    if args.search:
        params['keyword'] = args.search
    if args.sub_domain:
        params['sub_domain'] = args.sub_domain

    records = dnspod.get_records(**params)

    output("name", "type", "line", "value", "mx", "ttl", "status")
    output("-" * 50)
    for record in records:
        output(record['name'], record['type'], record['line'], record['value'],
               (record['mx'] if int(record['mx']) else '-'), record['ttl'],
               ('enabled' if int(record['enabled']) else 'disabled'))


@require_sa
def show(args):
    dnspod = DNSPod(domain=args.domain)
    if args.record_line:
        output('>>>', 'supported lines')
        for line in dnspod.get_record_lines():
            output(line)
    if args.record_type:
        output('>>>', 'supported types')
        for type_ in dnspod.get_record_types():
            output(type_)
    if args.user_log:
        output('>>>', 'user log')
        for log_line in dnspod.get_user_log():
            output(log_line)


@require_sa
def dump(args):
    """
    ref: http://cr.yp.to/djbdns/tinydns-data.html
    # record type:
    'A' : =img0.exampleio.com:211.147.4.31:300; +movie.example.com:211.147.4.31:300
    'CNAME' : Cimg1.exampleio.com:img1-exampleio-com.b0.aicdn.com.:86400
    'MX' : @example.com::aspmx.l.google.com.:10:ttl
    'NS' : .example.com::ns1.example.com.:300; &example.com::ns1.example.com.:300
    'TXT' : 'read.example.com:v=spf1 include\072mailgun.org ~all:86400
    'GENERIC' : :fqdn:dns_type:rdata:ttl:timestamp:lo (it starts with a colon)
    'AAAA' : :ipv6.example.com:28:\052\013\103\100\035\000\000\000\200\000\000\000\000\000\276\357:300
    """

    dnspod = DNSPod(domain=args.domain)
    records = dnspod.get_records()
    records.sort(reverse=True, key=lambda x: x['id'])

    for record in records:
        name, type_, value, mx, ttl, enable = record['name'], record['type'], record['value'], \
            record['mx'], record['ttl'], int(record['enabled'])
        domain = args.domain if name == '@' else '.'.join([name, args.domain])

        if type_ == 'A':
            if 'mail' in name:
                line = DNSPOD_TO_TINYDNS[type_][0].format(domain=domain, value=value, ttl=ttl)
            else:
                line = DNSPOD_TO_TINYDNS[type_][1].format(domain=domain, value=value, ttl=ttl)
        elif type_ == 'AAAA':
            line = DNSPOD_TO_TINYDNS[type_].format(domain=domain, value=ipv6_addr_to_tinydns_generic(value), ttl=ttl)
        elif type_ == 'MX':
            line = DNSPOD_TO_TINYDNS[type_].format(domain=domain, value=value, dist=mx, ttl=ttl)
        elif type_ == 'TXT':
            line = DNSPOD_TO_TINYDNS[type_].format(domain=domain, value=value.replace(':', '\\072'), ttl=ttl)
        else:
            line = DNSPOD_TO_TINYDNS[type_].format(domain=domain, value=value, ttl=ttl)

        if not enable:
            line = '#' + line

        output(line)


@require_sa
def ensure(args):
    """sub_domain -> (type, value, ttl, enable)"""
    dnspod = DNSPod(dry_run=args.dry_run, domain=args.domain)
    sub_domains = parse_sub_domains(args.sub_domains, args.domain)

    records = []
    for sub_domain in sub_domains:
        full_domain = sub_domain.rstrip('.') + '.' + DOMAIN
        value = args.value.format(domain=full_domain, sub_domain=sub_domain, type=args.type)
        sub_domain_records, msg = dnspod.add_or_modify_record(sub_domain, args.type, value,
                                                              ttl=args.ttl, mx=args.mx,
                                                              line=args.line,
                                                              force_enable=args.enable)
        if msg:
            logger.error('ensure sub-domain {sub_domain} failed: {msg}'
                         .format(sub_domain=sub_domain, msg=msg))
            exit(2)

        records += sub_domain_records

        if args.excl:
            dnspod.set_status(sub_domain, args.type, enable=False, except_values=[value])

    logger.info('sub-domains {sub_domains} ensured.'.format(sub_domains=sub_domains))


@require_sa
def remove(args):
    dnspod = DNSPod(dry_run=args.dry_run, domain=args.domain)
    sub_domains = parse_sub_domains(args.sub_domains, args.domain)

    records = []
    sub_domains_removed = []
    for sub_domain in sub_domains:
        full_domain = sub_domain.rstrip('.') + '.' + DOMAIN
        value = args.value.format(domain=full_domain, sub_domain=sub_domain, type=args.type)
        sub_domain_records, msg = dnspod.remove_records(sub_domain, args.type, value, args.line)
        if msg:
            logger.error('remove records of sub-domain ({sub_domain}, {type}, {value}) '
                         'failed: {msg}'
                         .format(sub_domain=sub_domain, msg=msg,
                                 type=type,
                                 value=value))
            exit(2)
        if sub_domain_records:
            sub_domains_removed.append(sub_domain)
        else:
            logger.warn('No records matched.')

        records += sub_domain_records

    logger.info('records of sub-domains ({sub_domains}, {type}, {value}) removed.'
                .format(sub_domains=sub_domains_removed,
                        type=args.type,
                        value=args.value))


def main(args=None):
    """
    e.g.
    $ sa-dns ensure sub1.subdomain1 --type CNAME --value {domain}.anotherdomain. --enable
    $ sa-dns remove subdomain1,subdomain2
    $ sa-dns -d example.com dump
    $ sa-dns show -tlu
    """
    parser = argparse.ArgumentParser(epilog=main.__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-u', '--user', help='LDAP username, use your OS user name by default.')
    parser.add_argument('-d', '--domain', default=DOMAIN,
                        help='Top level domain. (default: %(default)s)')

    # compatible with py2 & 3
    subparsers = parser.add_subparsers(help='Sub commands', dest='subparser')
    subparsers.required = True

    dump_parser = subparsers.add_parser('dump', help='Dump dns records.')
    dump_parser.set_defaults(func=dump)
    dump_parser.set_defaults(parser_name='dump')

    list_parser = subparsers.add_parser('list', help='List dns records.')
    list_parser.add_argument('-s', '--search', help='Filter by search keywords.')
    list_parser.add_argument('-S', '--sub-domain', help='Filter by sub domain.')
    list_parser.set_defaults(func=list_)
    list_parser.set_defaults(parser_name='list')

    show_parser = subparsers.add_parser('show', help='Show information.')
    show_parser.add_argument('-t', '--record-type', action='store_true',
                             help='Show record type supported')
    show_parser.add_argument('-l', '--record-line', action='store_true',
                             help='Show record line supported')
    show_parser.add_argument('-u', '--user-log', action='store_true',
                             help='Show user log')
    show_parser.set_defaults(func=show)
    show_parser.set_defaults(parser_name='show')

    ensure_parser = subparsers.add_parser('ensure',
                                          help="Add or modified, "
                                               "ensure the dns records that demanded to present.")
    ensure_parser.add_argument('sub_domains', help='Sub domains, comma-separated.')
    ensure_parser.add_argument('--type', choices=['A', 'CNAME', 'MX', 'TXT'],
                               required=True,
                               help='DNS record type.')
    ensure_parser.add_argument('--value',
                               required=True,
                               help='DNS record value. '
                                    'Support pattern, like `{domain}.h1.aqb.so`. '
                                    'Avaliable variable are domain, sub_domain, type')
    ensure_parser.add_argument('--ttl', default=DEFAULT_TTL, help='DNS record time-to-live.'
                                                                  '(default: %(default)s)')
    ensure_parser.add_argument('--mx', default=DEFAULT_MX,
                               help='MX priority value, required by MX record.'
                                    '(default: %(default)s)')
    ensure_parser.add_argument('--line', default=DEFAULT_LINE, help='DNS record line.'
                                                                    '(default: %(default)s)')
    ensure_status = ensure_parser.add_mutually_exclusive_group(required=True)
    ensure_status.add_argument('--enable', action='store_true', help='Enable the DNS records.')
    ensure_status.add_argument('--disable', action='store_true', help='Disable the DNS records.')
    ensure_parser.add_argument('--excl', '--exclusive', action='store_true',
                               help='Exclusive, disable other records with the same type.')
    ensure_parser.add_argument('--dry-run', action='store_true', help='Dry run.')
    ensure_parser.set_defaults(func=ensure)
    ensure_parser.set_defaults(parser_name='ensure')

    remove_parser = subparsers.add_parser('remove', help='Remove dns records.')
    remove_parser.add_argument('sub_domains', help='Sub domains, comma-separated.')
    remove_parser.add_argument('--type', choices=['', 'A', 'CNAME', 'MX', 'TXT'],
                               default='',
                               help='DNS record type.')
    remove_parser.add_argument('--value',
                               default='',
                               help='DNS record value. '
                                    'Support pattern, like `{domain}.h1.aqb.so`. '
                                    'Avaliable variable are domain, sub_domain, type')
    remove_parser.add_argument('--line', default=DEFAULT_LINE, help='DNS record line.'
                                                                    '(default: %(default)s)')
    remove_parser.add_argument('--dry-run', action='store_true', help='Dry run.')
    remove_parser.set_defaults(func=remove)
    remove_parser.set_defaults(parser_name='remove')

    monitor_parser = subparsers.add_parser('monitor', help='DNSPod Monitor.')
    monitor_subparser = monitor_parser.add_subparsers(help='Monitor Sub Commands.')
    monitor_list_parser = monitor_subparser.add_parser('list', help='List monitors.')
    monitor_list_parser.add_argument('-a', '--all', action='store_true', help='List all monitors.')
    monitor_list_parser.add_argument('-S', '--sub-domain', help='Filter by sub domain.')
    monitor_list_parser.set_defaults(func=list_monitor)
    monitor_add_parser = monitor_subparser.add_parser('add', help='Add monitors.')
    monitor_add_parser.add_argument('sub_domains', help='Sub domains, comma-separated.')
    monitor_add_parser.add_argument('-p', '--port', default=80, help='monitor HTTP(S) port.(default: %(default)s)')
    monitor_add_parser.add_argument('-i', '--monitor-interval', default=60, choices=[60, 180, 360, 600],
                                    help='monitor interval. (default: %(default)s)')
    monitor_add_parser.add_argument('-t', '--monitor-schema-type', default='http', choices=['http', 'https'],
                                    help='monitor schema type. (default: %(default)s)')
    monitor_add_parser.add_argument('-P', '--monitor-uri-path', default='/', help='monitor uri path. (default: %(default)s)')
    monitor_add_parser.add_argument('--points', default='ctc,cuc,cmc,ctc-2,cuc-2,cmc-2,ctc-3,cuc-3,cmc-3',
                                    help='comma-separated monitor points/probes. (default: %(default)s)')
    monitor_add_parser.add_argument('-b', '--bak-ip-mode', default='pass',
                                    help='chose from pass, pause, pause2, auto or comma-separated ip. (default: %(default)s)')
    monitor_add_parser.add_argument('--send-to-callback', action='store_true', help='whether send to dns monitor callback')
    monitor_add_parser.add_argument('--dry-run', action='store_true', help='Dry run.')
    monitor_add_parser.set_defaults(func=add_monitor)
    monitor_remove_parser = monitor_subparser.add_parser('remove', help='Remove monitors.')
    monitor_remove_parser.add_argument('sub_domains', help='Sub domains, comma-separated.')
    monitor_remove_parser.add_argument('--dry-run', action='store_true', help='Dry run.')
    monitor_remove_parser.set_defaults(func=remove_monitor)

    args = parser.parse_args(args)
    args.user = args.user or get_os_username()

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    args.func(args)



# coding: utf-8

import shlex
import importlib

from sa_tools_core.libs.process import Process


class Client(Process):
    def __init__(self, cmds=None, func=None):
        super(Client, self).__init__(cmds)
        self.func = func

    def __str__(self):
        if self.func:
            func_str = '%s:%s' % (self.func.__module__, self.func.__name__)
        else:
            func_str = 'None'
        return '%s(cmds=%s, func=%s)' % (self.__class__.__name__, self.cmds, func_str)

    __repr__ = __str__

    def __getattr__(self, name):
        if not self.func:
            mod = importlib.import_module('sa_tools_core.%s' % name)
            return self.__class__(func=mod.main)

        return super(Client, self).__getattr__(name)

    def bake(self, *a, **kw):
        cmds = list(self.cmds)
        proc = Client(cmds, self.func)
        proc._parse_args(*a, **kw)
        return proc

    def call(self, cmdstr='', env=None):
        extra_cmds = shlex.split(cmdstr)
        return self.func(self.cmds + extra_cmds)


if __name__ == '__main__':
    c = Client()
    print(c.notify(wework='user1', content='hehe'))
    print(c.uptime())
    print(c.notify.test)
    print(c.dns.list(S='@'))



import argparse

from sa_tools_core.libs.github import GithubRepo, GITHUB_API_ENTRYPOINT, github_secret_func
# ...

def operate_collaborator(args):
    repo = GithubRepo(args.org, args.repo, entrypoint=GITHUB_API_ENTRYPOINT, secret_func=github_secret_func)
    if args.action == 'add':
        repo.ensure_collaborator(args.username, permission=args.permission)
    if args.action == 'remove':
        repo.remove_collaborator(args.username)

def main():
    parser = argparse.ArgumentParser(
        description="operator github or ghe using cli"
    )

    subparsers = parser.add_subparsers(help='Sub commands', dest='subparser')
    subparsers.required = True

    collaborator = subparsers.add_parser('collaborator')


    collaborator.add_argument(
        "--org",
        help="github organization",
        required=True
    )
    collaborator.add_argument(
        "--repo",
        help="github repo name",
        required=True
    )
    collaborator.add_argument(
        "action",
        choices=["add", "remove"],
        help="action you want to take, add or remove a collaborator"
    )
    collaborator.add_argument(
        "--username",
        required=True,
        help="github username"
    )
    collaborator.add_argument(
        "--permission",
        default="push",
        choices=["pull", "push", "admin", "maintain", "triage"]
    )

    collaborator.set_defaults(func=operate_collaborator)

    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()


# coding: utf-8

import time
from datetime import datetime


# TODO: more flexiable time format
def timeformat(time_string):
    now_ts = int(time.time())
    dt = None
    try:
        time_int = int(time_string)
        if time_int < 0:
            ts = now_ts + time_int * 60
            dt = datetime.fromtimestamp(ts)
            return dt
        else:
            # timestamp ? time range limit ?
            return None
    except:
        pass
    for p in ('%Y-%m-%d %H:%M',
              '%Y-%m-%d %H:%M:%S',
              '%Y-%m-%dT%H:%M:%S',
              ):
        try:
            dt = datetime.strptime(time_string, p)
            return dt
        except:
            pass
    return dt



# coding: utf-8

from __future__ import absolute_import

import logging

from wechatpy.enterprise import WeChatClient

from sa_tools_core.utils import get_config

logger = logging.getLogger(__name__)

# CORPID = 'aaaaaaaaaaaaaaaaaa'
# CORPSECRET = "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
# APPID = 100002 or 2


def send_message(
    touser=None,
    content=None,
    msg_type="text",
    toparty=None,
    totag=None,
    **kwargs,
):
    secret = get_config("wework")
    CORPID, CORPSECRET, APPID = secret.split(":")
    agentid = int(APPID)
    api = WeChatClient(CORPID, CORPSECRET)
    message = dict(
        agent_id=agentid,
        user_ids=touser,
        content=content,
        party_ids=toparty,
        tag_ids=totag,
    )

    if msg_type == "text":
        ret = api.message.send_text(**message)
    elif msg_type == "markdown":
        ret = api.message.send_markdown(**message)
    else:
        raise Exception("unsupported message type")
    return ret



# coding: utf-8

from __future__ import print_function

import os

PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))

IPMI_TIMEOUT = 2
IPMI_RETRIES = 1

########################
# CONFIGURATION SECTION
########################

CONFIG_DIR = '/etc/sa-tools/'
EXTERNAL_DOMAINS_CONFIG_FILE = CONFIG_DIR + '/external_domains'

#######################################
# # Notification (sa-notify related)
#######################################

# sa-node related, not used yet
NODE_IRC_CHANNEL = '#sysadmin'
NODE_SLACK_CHANNEL = '#sa'
NODE_MAIL_TO = 'sa@example.com'

# default mail from addr
SYSADMIN_EMAIL = 'sysadmin@example.com'

# SMTP server related
SMTP_SERVER = ''
# If port is zero, the standard port is used.
SMTP_SERVER_PORT = 0
# Indicate the config file name for SMTP credentials, e.g. /etc/sa-tools/email
# SMTP_CREDENTIALS_CONFIG = 'email'
SMTP_CREDENTIALS_CONFIG = ''
SMTP_SSL = False

# proxies for slack, etc.
PROXIES = {
    "http": "http://gfw:2333",
    "https": "http://gfw:2333",
}

DEFAULT_LARK_TENANT = "your_company"

########################
# # Elasticsearch
########################

SA_ES_HOSTS = ['es.svc:8080']
SA_ES_VERSION = (7, 5, 1)
SA_ES_USER = ''
SA_ES_PASSWD = ''
SA_ES_NGINX_ACCESS_INDEX_PREFIX = 'heka-nginx.access-'
SA_ES_NGINX_ACCESS_INDEX_TIME_FORMAT = '%Y.%m.%d'
SA_ES_NGINX_ACCESS_TIMESTAMP_FIELD_NAME = "Timestamp"
SA_ES_NGINX_ACCESS_LOG_FIELD_NAME = "Payload"

########################
# # DNS
########################

# can be a string pattern which contains {cb_token}
DNS_MONITOR_CALLBACK_URL = 'http://example.com/api/callback/dnsmonitor/{cb_token}'
DEFAULT_DNS_DOMAIN = 'example.com'

########################
# # Scripts
########################

ANSIBLE_INVENTORY_CONFIG_PATH = ['/etc/ansible/hosts']
ANSIBLE_MODULE_PATH = []

########################
# # Disk
########################

NCDU_EXPORT_DATA_PATH = "/data1"
NCDU_JOB_LOCK_PREFIX = "/tmp/sa-disk-ncdu-lock"

########################
# # Icinga2
########################

ICINGA_EMAIL = 'icinga@example.com'
# see https://docs.sentry.io/error-reporting/configuration/?platform=python for more details
# SENTRY_DSN = 'https://<key>@sentry.io/<project>'
SENTRY_DSN = ''

# icinga alert page base url, every notify sent by sa-icinga will be attrtched with a wiki link
# For example : http://yourwiki.com/alerts/sshd_down
# Set to '' if you don't want it.
ALERT_WIKI_BASE_URL = ''

# notification gateway is a web service that stores and analyizes notifications. We have one in douban, you can build
# your own.
NOTIFICATION_GATEWAY_API = ''
NOTIFICATION_GATEWAY_TIMEOUT = 10  # 10s

ICINGA_CACERT = '/etc/icinga2/ssl/certs/ca.pem'
# you need to inhert IcingaClusterConfig and impl your own config class
# see sa_tools_core.libs.icinga for more details
ICINGA_CLUSTER_CONFIG_CLASS = 'sa_tools_core.libs.icinga:IcingaClusterConfig'

########################
# # BS(Black Stone)
########################

BS_CMD_PATTERN = 'qcloudcli "{module}" "{action}" {params}'
BS_DEFAULT_ATTRS = (
    'alias',
    'eipName',
    'subnetName',
    # 'vpcName', 'natName', 'instanceAlias',
    # 'instanceId', 'vlanId', 'natUid', 'unInstanceId', 'subnetId',
    'lanIp',
    'vpcIp',
    'eip',
    'cidrBlock',
    'cidr',
)
BS_DEFAULT_PARAMS = {
    'limit': 100,
}
BS_DEFAULT_PARAMS_BM = {
    'vpcId': 1001,
    'unVpcId': 'vpc-xxxxxxxx',
    'zoneId': 1000800001,  # 可用区ID。可通过 DescribeRegions 接口用来获取黑石物理机可用区。
}
BS_PLURAL_SUFFIX = ['s', 'List', 'Set']

########################
# # TC(Tencent CLI)
########################

TENCENT_DEFAULT_REGIN = 'ap-beijing'
TENCENT_DEFAULT_PARAMS = {
    'limit': 10,
    'offset': 0,
    'enhanced_service_security_service_enabled': False,
    'enhanced_service_monitor_service_enabled': False,
}

########################
# # Others
########################

GITHUB_API_ENTRYPOINT = 'https://ghe.yourdomain.com/api/v3'

try:
    from local_config import *  # NOQA
except Exception:
    pass

try:
    from io import open
    config_path = os.path.join(CONFIG_DIR, "config.py")
    exec(open(config_path, encoding='utf-8').read())
except Exception:
    print('WARNING: failed to load config %s.' % config_path)



# coding: utf-8
'''
An ansible wrapper which makes ansible 2.x be compatible with ansible 1.9

refs:
    - https://docs.ansible.com/ansible/latest/dev_guide/developing_api.html
    - https://github.com/ansible/ansible/blob/1da5e21289d7407925190139165a9cd7a3f69960/examples/scripts/uptime.py
'''

from __future__ import absolute_import, print_function

import logging
import ansible

ansible_version = ansible.__version__.split('.')  # NOQA

# if ansible_version < (2, 0, 0):
#     import ansible.utils.template  # NOQA
#     from ansible import errors  # NOQA
#     from ansible.callbacks import DefaultRunnerCallbacks  # NOQA
#     from ansible.inventory import Inventory  # NOQA
#     from ansible.runner import Runner  # NOQA
# else:

from ansible.module_utils.common.collections import ImmutableDict
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager
from ansible.inventory.manager import InventoryManager
from ansible.inventory.host import Host
from ansible.playbook.play import Play
from ansible.executor.task_queue_manager import TaskQueueManager
from ansible.plugins.callback import CallbackBase
from ansible import context
import ansible.constants as C  # NOQA

from sa_tools_core.consts import ANSIBLE_INVENTORY_CONFIG_PATH, ANSIBLE_MODULE_PATH

logger = logging.getLogger(__name__)

loader = DataLoader()  # Takes care of finding and reading yaml, json and ini files

# create inventory, use path to host config file as source or hosts in a comma separated string
inventory = InventoryManager(loader=loader, sources=ANSIBLE_INVENTORY_CONFIG_PATH)

Inventory = lambda: inventory  # NOQA

# variable manager takes care of merging all the different sources
# to give you a unified view of variables available in each context
variable_manager = VariableManager(loader=loader, inventory=inventory)


class DefaultRunnerCallbacks(CallbackBase):
    '''
    make ansible 2.x Callback be compatible with ansible 1.9
    '''
    def __init__(self):
        self.results = {'contacted': {}, 'dark': {}}
        super(DefaultRunnerCallbacks, self).__init__()

    def v2_runner_on_failed(self, result, ignore_errors=False):
        super(DefaultRunnerCallbacks, self).v2_runner_on_failed(result, ignore_errors)
        host = result._host.get_name()
        self.results['contacted'][host] = result._result
        self.on_failed(host, result._result, ignore_errors)

    def v2_runner_on_ok(self, result):
        super(DefaultRunnerCallbacks, self).v2_runner_on_ok(result)
        host = result._host.get_name()
        self.results['contacted'][host] = result._result
        self.on_ok(host, result._result)

    def v2_runner_on_skipped(self, result):
        super(DefaultRunnerCallbacks, self).v2_runner_on_skipped(result)
        host = result._host.get_name()
        self.results['dark'][host] = result._result
        self.on_skipped(host, self._get_item_label(getattr(result._result, 'results', {})))

    def v2_runner_on_unreachable(self, result):
        super(DefaultRunnerCallbacks, self).v2_runner_on_unreachable(result)
        host = result._host.get_name()
        self.results['dark'][host] = result._result
        self.on_unreachable(host, result._result)

    def v2_playbook_on_no_hosts_matched(self):
        super(DefaultRunnerCallbacks, self).v2_playbook_on_no_hosts_matched()
        self.on_no_hosts()

    def on_failed(self, host, res, ignore_errors=False):
        pass

    def on_ok(self, host, res):
        pass

    def on_skipped(self, host, item=None):
        pass

    def on_unreachable(self, host, res):
        pass

    def on_no_hosts(self):
        pass


class Runner(object):
    def __init__(
        self,
        connection='ssh',
        module_path=None,  # ex: /usr/share/ansible
        module_name='shell',
        module_args=None,
        forks=C.DEFAULT_FORKS,  # parallelism level
        pattern=None,  # which hosts?  ex: 'all', 'acme.example.org'
        remote_user=C.DEFAULT_REMOTE_USER,  # ex: 'username'
        callbacks=None,  # used for output
        run_hosts=None,  # an optional list of pre-calculated hosts to run on, or host pattern
        become=False,  # whether to run privelege escalation or not
        become_method=C.DEFAULT_BECOME_METHOD,
    ):

        self.callback = callbacks

        # find hosts that match the pattern
        if not run_hosts:
            run_hosts = pattern

        if isinstance(run_hosts, list):
            run_hosts = [h.name if isinstance(h, Host) else h for h in run_hosts]

        # since the API is constructed for CLI it expects certain options to always be set in the context object
        context.CLIARGS = ImmutableDict(
            connection=connection,
            remote_user=remote_user,
            module_path=module_path or ANSIBLE_MODULE_PATH,
            forks=forks,
            become=become,
            become_method=become_method or 'sudo',
            check=False,
            diff=False,
            verbosity=0)
        self.forks = forks
        self.play = self._create_play(module_name, module_args, run_hosts)

    @classmethod
    def _create_play(cls, module_name, module_args, hosts):
        # create data structure that represents our play,
        # including tasks, this is basically what our YAML loader does internally.
        play_source = dict(
            name="Ansible Play",
            hosts=hosts,
            gather_facts='no',
            tasks=[
                dict(action=dict(module=module_name, args=module_args)),
            ])

        # Create play object, playbook objects use .load instead of init or new methods,
        # this will also automatically create the task objects from the info provided in play_source
        play = Play().load(play_source, variable_manager=variable_manager, loader=loader)
        return play

    def run(self):
        # Run it - instantiate task queue manager,
        # which takes care of forking and setting up all objects to iterate over host list and tasks
        tqm = None
        try:
            tqm = TaskQueueManager(
                inventory=inventory,
                variable_manager=variable_manager,
                loader=loader,
                passwords=dict(),
                stdout_callback=self.
                callback,  # Use our custom callback instead of the ``default`` callback plugin, which prints to stdout
                forks=self.forks,
            )
            # most interesting data for a play is actually sent to the callback's methods
            _ = tqm.run(self.play)
        except Exception:
            raise
        finally:
            # we always need to cleanup child procs and the structures we use to communicate with them
            if tqm is not None:
                tqm.cleanup()

            # Remove ansible tmpdir
            # shutil.rmtree(C.DEFAULT_LOCAL_TMP, True)
            if loader:
                loader.cleanup_all_tmp_files()

        return self.callback.results



# coding: utf-8

import json
import base64
import logging
from six.moves.urllib.parse import urlencode
from six.moves.urllib.request import urlopen, Request

from sa_tools_core.utils import get_config

PUSH_URL = 'https://api.pushbullet.com/v2/pushes'

logger = logging.getLogger(__name__)


def send_message(email, title, body):
    PUSH_KEY = get_config('pushbullet')

    data = {
        "email": email,
        "type": 'note',
        "title": title,
        "body": body,
    }
    payload = urlencode(data)
    req = Request(PUSH_URL)
    base64string = base64.encodestring('%s:%s' % (PUSH_KEY, ''))[:-1]
    authheader = "Basic %s" % base64string
    req.add_header("Authorization", authheader)

    resp = json.loads(urlopen(req, payload).read())
    if not resp['receiver_email'] == email:
        raise Exception('api failed, resp: %s' % resp)



# coding: utf-8

from __future__ import absolute_import

import logging
from functools import wraps

import sentry_sdk
from sentry_sdk.client import Client
from sentry_sdk.hub import Hub
from sentry_sdk.integrations.excepthook import ExcepthookIntegration
from sentry_sdk.integrations.dedupe import DedupeIntegration
from sentry_sdk.integrations.stdlib import StdlibIntegration
from sentry_sdk.integrations.modules import ModulesIntegration
from sentry_sdk.integrations.argv import ArgvIntegration

from sa_tools_core.consts import SENTRY_DSN

logger = logging.getLogger(__name__)

try:
    _client = Client(
        dsn=SENTRY_DSN,
        default_integrations=False,
        integrations=[
            ExcepthookIntegration(),
            DedupeIntegration(),
            StdlibIntegration(),
            ModulesIntegration(),
            ArgvIntegration(),
        ],
        max_breadcrumbs=5,
        attach_stacktrace=True,
    )
    _hub = Hub(_client)
except:
    logger.exception('failed to load sentry: ')


def report(msg=None, **kw):
    try:
        extra = kw.pop('extra', {})

        with sentry_sdk.push_scope() as scope:
            for k, v in extra.items():
                scope.set_extra(k, v)
            scope.level = kw.pop('level', logging.ERROR)

            if 'user' in kw:
                scope.user = kw.get('user')

            if msg:
                _hub.capture_message(msg, level=scope.level)
            else:
                _hub.capture_exception()
    except Exception:
        logger.exception('report to sentry failed: ')


def send_sentry(func):
    @wraps(func)
    def _(*a, **kw):
        try:
            return func(*a, **kw)
        except Exception:
            report()
            raise
    return _



# coding: utf-8

from __future__ import print_function

import sys
import logging
import argparse
from pprint import pprint

import six

from sa_tools_core.libs.permission import require_sa, require_user  # NOQA
from sa_tools_core.libs.es import ESQuery
from sa_tools_core.libs.timeformat import timeformat
from sa_tools_core.consts import (SA_ES_HOSTS, SA_ES_VERSION,
                                  SA_ES_NGINX_ACCESS_INDEX_PREFIX,
                                  SA_ES_NGINX_ACCESS_INDEX_TIME_FORMAT,
                                  SA_ES_NGINX_ACCESS_TIMESTAMP_FIELD_NAME,
                                  SA_ES_NGINX_ACCESS_LOG_FIELD_NAME)
try:
    from sa_tools_core.consts import (SA_ES_USER, SA_ES_PASSWD)
except Exception as e:
    SA_ES_USER = ''
    SA_ES_PASSWD = ''
from sa_tools_core.utils import get_os_username, props, i2ip

logger = logging.getLogger(__name__)
VERBOSE_LEVEL = (
    (0, logging.WARN),
    (1, logging.INFO),
    (2, logging.DEBUG),
)

N_DOC_DEFAULT = 10


class NginxAccessESQuery(ESQuery):
    def __init__(self,
                 es_hosts=SA_ES_HOSTS,
                 index_prefix=SA_ES_NGINX_ACCESS_INDEX_PREFIX,
                 timestamp_field=SA_ES_NGINX_ACCESS_TIMESTAMP_FIELD_NAME,
                 index_time_format=SA_ES_NGINX_ACCESS_INDEX_TIME_FORMAT,
                 es_user=SA_ES_USER,
                 es_passwd=SA_ES_PASSWD):
        super(NginxAccessESQuery, self).__init__(es_hosts=es_hosts, index_prefix=index_prefix,
                                                 timestamp_field=timestamp_field,
                                                 index_time_format=index_time_format,
                                                 es_user=es_user, es_passwd=es_passwd)

    def make_body(self, time_range=None, query_string=None, term_dict=None,
                  aggregations=None, sort=None,
                  **kw):
        must_list = [{"range": time_range}]
        if term_dict:
            must_list += [{"term": {k: v}} for k, v in term_dict.items()]
        body = {
            "query": {
                "bool": {
                    "must": {
                        "query_string": query_string,
                    },
                    "filter": {
                        "bool": {
                            "must": must_list,
                            # "must_not": [],
                        }
                    }
                }
            },
            "aggs": aggregations,
            "sort": sort,
        }
        logger.debug('make_body: %s' % body)
        return body

    def query(self, lucene_query_string="*", term_dict=None, aggregations=None, sort=None,
              start=-15, duration=15, size=0):
        return super(NginxAccessESQuery, self).query(lucene_query_string=lucene_query_string,
                                                     term_dict=term_dict,
                                                     aggregations=aggregations, sort=sort,
                                                     start=start, duration=duration, size=size)


class AggregationsUtils(object):
    @staticmethod
    def by(x, y=None, y_script=None, size=20):
        """ x by y. e.g. by(bandwidth, 'remote_addr') """
        aggs_name = 'my_aggs'
        if not y and not y_script:
            return {aggs_name: x} if x != '_count' else {}
        order_dir = 'desc'
        x_name = str(x) if isinstance(x, str) else 'sub_aggs'
        order_name = x_name
        if isinstance(x, dict) and 'percentiles' in x:
            order_name = '%s.%s' % (x_name, x['percentiles']['percents'][0])
        if isinstance(y, str):
            aggs = dict(terms=dict(field=y,
                                   size=size,
                                   order={order_name: order_dir}))
        elif y_script:
            aggs = dict(terms=dict(script=y_script,
                                   size=size,
                                   order={order_name: order_dir}))
        else:  # dictify
            aggs = y
        if x != '_count':
            aggs['aggs'] = {x_name: x}
        return {aggs_name: aggs}

    # x
    class X(object):
        @staticmethod
        def _parse(*a):
            if a[0] in builtin_x:
                x = getattr(A.X, a[0])
                # Get the function out of unbound method.
                # In Python 3, unbound methods don't exist, so this function
                # just returns the function object unchanged.
                return six.get_unbound_function(x)(*[A.Y._alias_field(f) for f in a[1:]]) if callable(x) else x
            else:
                return six.get_unbound_function(A.X._)(*a)

        _ = lambda agg_func, field: {agg_func: dict(field=field)}  # NOQA
        qps = count = '_count'
        bandwidth = _('sum', 'bytes_sent')
        sum_response_time = _('sum', 'upstream_response_time')
        avg_response_time = _('avg', 'upstream_response_time')
        percentiles = lambda field_name, percents: {'percentiles': dict(field=field_name, percents=percents)}  # NOQA
        mean_response_time = percentiles('upstream_response_time', [50])
        unique_count = lambda field: {'cardinality': dict(field=field)}  # NOQA
        unique_ip_count = unique_count('remote_addr')
        unique_bid_count = unique_count('browser_id')

    # y
    class Y(object):
        @staticmethod
        def _parse(*a):
            if a[0] in builtin_y:
                y = getattr(A.Y, a[0])
                return six.get_unbound_function(y)(*a[1:]) if callable(y) else y
            else:
                return a[0]

        @staticmethod
        def _alias_field(alias):
            if alias in builtin_y and isinstance(getattr(A.Y, alias), str):
                return getattr(A.Y, alias)
            return alias

        ip = 'remote_addr'
        ua = 'http_user_agent'
        uid = 'user_id'
        bid = 'browser_id'
        nurl = 'normalize_url'
        dt = date_histogram = lambda interval: \
            {'date_histogram': dict(field=SA_ES_NGINX_ACCESS_TIMESTAMP_FIELD_NAME,
                                    interval=interval, time_zone="Asia/Shanghai")}
        date_histogram_1s = date_histogram('1s')


A = AggregationsUtils
builtin_x = props(A.X)
builtin_y = props(A.Y)


def _query(args, es, start, doc_number=None):
    if args.term:
        term_dict = {A.Y._alias_field(args.term[i << 1]): args.term[i << 1 | 1]
                     for i in range(len(args.term) >> 1)}
    else:
        term_dict = None

    if args.agg_x:
        aggs_params = {'x': A.X._parse(*args.agg_x), 'size': args.agg_number}
        if args.agg_y:
            aggs_params['y'] = A.Y._parse(*args.agg_y)
        elif args.agg_multi_y:
            aggs_params['y_script'] = " + ' ' + ".join("doc['%s'].value" % A.Y._alias_field(field)
                                                       for field in args.agg_multi_y)
        elif args.agg_script_y:
            aggs_params['y_script'] = args.agg_script_y
        aggs = A.by(**aggs_params)
        doc_number = 0 if doc_number is None else doc_number
        sort = None
    else:
        aggs = None
        doc_number = N_DOC_DEFAULT if doc_number is None else doc_number
        sort = [{SA_ES_NGINX_ACCESS_TIMESTAMP_FIELD_NAME: {'order': "desc",
                                                           'unmapped_type': "boolean"}}]

    res = es.query(term_dict=term_dict,
                   aggregations=aggs,
                   # aggregations=A.by(A.X.bandwidth, A.Y.ip),
                   # aggregations=A.by(A.X.unique_bid_count, A.Y.ip, size=args.agg_number),
                   # aggregations=A.by(A.X.unique_bid_count),
                   # aggregations=A.by(A.X.qps, A.Y.date_histogram_1s),
                   sort=sort,
                   start=start,
                   duration=args.duration,
                   size=doc_number,
                   lucene_query_string=args.query_string)

    return res


def parse_es_result(args, res):
    def as_string(s):
        if SA_ES_VERSION < (3, 0, 0):
            # the following code is for old elasticsearch, to translate ip format
            if args.agg_multi_y:
                remote_addr_indexes = {i for i, field in enumerate(args.agg_multi_y)
                                       if A.Y._alias_field(field) == 'remote_addr'}
                sep = ' '
                parts = s.split(sep)
                s = sep.join(i2ip(p) if i in remote_addr_indexes else p for i, p in enumerate(parts))
        return s

    ret = {}
    if 'hits' in res.get('hits', {}):
        ret['docs'] = [doc.get('_source', {}).get(SA_ES_NGINX_ACCESS_LOG_FIELD_NAME) for doc in res['hits']['hits']]
    if 'aggregations' in res:
        my_aggs = res['aggregations'].get('my_aggs', {})
        buckets = my_aggs.get('buckets', [])
        if buckets:
            possible_x_names = [k for k, v in buckets[0].items()
                                if isinstance(v, dict) and 'value' in v]
            x_name = possible_x_names[0] if possible_x_names else 'doc_count'
            ret['aggs'] = [(as_string(b.get('key_as_string') or b.get('key')),
                            b['doc_count'] if x_name == 'doc_count'
                            else b.get(x_name).get('value'))
                           for b in buckets]
        elif 'value' in my_aggs:
            ret['aggs'] = [('', my_aggs['value'])]

    return ret


@require_user
def query(args):
    es = NginxAccessESQuery()

    res = _query(args, es, args.start, doc_number=args.number)

    if args.raw:
        pprint(res.pop(u'aggregations', None))
        pprint(res)
    else:
        ret = parse_es_result(args, res)
        if 'docs' in ret and ret['docs']:
            print('>>', 'docs')
            for i in ret['docs']:
                print(i)
        if 'aggs' in ret:
            print('>>', 'aggs')
            for k, v in ret['aggs']:
                if args.min_value is None or float(v) >= float(args.min_value):
                    print(k, v)


@require_user
def analyze(args):
    es = NginxAccessESQuery()
    res = (_query(args, es, args.start_normal), _query(args, es, args.start_abnormal))
    if args.raw:
        pprint(res)
        return 0

    ret = [parse_es_result(args, r).get('aggs', {}) for r in res]
    dicts = [dict(rt) for rt in ret]

    for k, v in ret[1]:
        if k in dicts[0]:
            print(k, 'before', dicts[0][k], 'now', v)
        else:
            print(k, v, 'new')
    for k, v in ret[0]:
        if k not in dicts[1]:
            print(k, v, 'disappear')


@require_user
def explain(args):
    es = NginxAccessESQuery()
    res = es.get_mapping(fields=args.field)
    pprint(res)


def main():
    """
    e.g.

    ## query

    ### normal query
    $ sa-access query --term ip 1.1.1.1
    $ sa-access query -q 2.2.2.2 --term appname app1 -n 20
    $ sa-access query --term appname app1 remote_addr 2.2.2.2 -n 10
    $ sa-access query -q 'remote_addr:[1.1.1.0 TO 1.1.1.254]' -n10
    $ sa-access query -q 'remote_addr:"1.1.1.0/24"' -n10

    ### aggs query
    $ sa-access query -x bandwidth
    $ sa-access query -x sum bytes_sent
    $ sa-access query -t host movie.example.com -x bandwidth --by dt 30s
    $ sa-access query -t host movie.example.com -x max bytes_sent --by dt 30s
    $ sa-access query -q 'Baiduspider' -t host movie.example.com -x bandwidth --by nurl
    $ sa-access query -t host movie.example.com -x bandwidth --by ip
    $ sa-access query -x count --by ip
    $ sa-access query -x unique_count browser_id --by ip

    #### by multi field or script
    $ sa-access query --term appname app1 -x count --by-multi remote_addr normalize_url
    $ sa-access query --term appname app1 -x count \
            --by-script "doc['remote_addr'].value + ' ' + doc['normalize_url'].value"
    ##### by ip and nurl
    $ sa-access query --term appname app1 -x count \
            --by-script "def ip=doc['remote_addr'].value; \
            ip + ' ' + doc['normalize_url'].value"
    ###### If you use old version elasticsearch, you will get ip in long integer, then you can do the following:
    $ sa-access query --term appname app1 -x count \
            --by-script "def ip=doc['remote_addr'].value;(ip >>24) \
            + '.' + ((ip >> 16) % 256) + '.' + ((ip >> 8) % 256) \
            + '.' + (ip % 256) + ' ' + doc['normalize_url'].value"
    ##### by ip section and nurl
    $ sa-access query --term appname app1 -x count \
            --by-script "def ip=doc['remote_addr'].value; \
            ip.substring(0, ip.lastIndexOf('.')) + ' ' + doc['normalize_url'].value"

    ## analyze

    ### qps is abnormal, we try to find which (ip, nurl) pair contribute to it.
    $ sa-access analyze --term appname app1 -x count --by-multi remote_addr nurl

    $ sa-access analyze --term appname app1 -x bandwidth --by-multi nurl

    $ sa-access analyze --term host music.example.com -x count --by nurl \
            -a '2017-03-28 09:30' -d 15 -b '2017-03-28 10:30'

    ## explain

    $ sa-access explain
    $ sa-access explain -f remote_addr upstream_response_time

    """
    parser = argparse.ArgumentParser(epilog=main.__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-u', '--user', help='LDAP username, use your OS user name by default.')
    parser.add_argument('-v', '--verbose', action='count', default=0, help='verbose.')

    # compatible with py2 & 3
    subparsers = parser.add_subparsers(help='Sub commands', dest='subparser')
    subparsers.required = True

    explain_parser = subparsers.add_parser('explain', help='show the explanation of fields')
    explain_parser.add_argument('-f', '--field', nargs='*', help='es doc field')
    explain_parser.set_defaults(func=explain)

    # TODO: support more time format

    # sub command, diff tops
    analyze_parser = subparsers.add_parser('analyze', help='analyze access log to find '
                                                           'abnormal IP, UA, UID, NURL, etc.')
    analyze_parser.add_argument('-r', '--raw', action='store_true',
                                help='Whether to use raw json output.')
    analyze_parser.add_argument('-q', '--query-string', help='Lucene query string.')
    analyze_parser.add_argument('-t', '--term', nargs='*',
                                metavar='<term name> <term value>',
                                help='term query.')
    # time
    analyze_parser.add_argument('-a', '--start-normal', type=timeformat, default=timeformat(-30),
                                help='start time of the normal period(default: %(default)s '
                                     'means %(default)s mins before, or specify datetime string).')
    analyze_parser.add_argument('-b', '--start-abnormal', type=timeformat, default=timeformat(-15),
                                help='start time of the abnormal period(default: %(default)s '
                                     'means %(default)s mins before, or specify datetime string).')
    analyze_parser.add_argument('-d', '--duration', type=int, default=15,
                                help='time duration(default: %(default)s '
                                     'means a duration of %(default)s mins).')
    # aggs
    analyze_parser.add_argument('-N', '--agg-number', type=int, default=20,
                                help='size of agg y(default: %(default)s).')
    analyze_parser.add_argument('-x', '--agg-x', nargs='*', required=True,
                                help='x of the aggregation(built-in: %s, '
                                     'or custom: `<agg_func> <field>`).' % builtin_x)
    analyze_parser.add_argument('-y', '--agg-y', '--by', nargs='*',
                                help='y of the aggregation(built-in: %s, '
                                     'or custom: `<field>`).' % builtin_y)
    analyze_parser.add_argument('-m', '--agg-multi-y', '--by-multi', nargs='*',
                                help='y of the aggregation, multi field')
    analyze_parser.add_argument('-S', '--agg-script-y', '--by-script',
                                help='y of the aggregation, script')
    analyze_parser.set_defaults(func=analyze)

    query_parser = subparsers.add_parser('query', help='query nginx access')
    query_parser.add_argument('-r', '--raw', action='store_true',
                              help='Whether to use raw json output.')
    query_parser.add_argument('-q', '--query-string', help='Lucene query string.')
    query_parser.add_argument('-s', '--start', type=timeformat, default=timeformat(-15),
                              help='start time(default: %(default)s '
                                   'means %(default)s mins before, or specify datetime string).')
    query_parser.add_argument('-d', '--duration', type=int, default=15,
                              help='time duration(default: %(default)s '
                                   'means a duration of %(default)s mins).')
    query_parser.add_argument('-n', '--number', type=int,
                              help='number of documents(default: 0 for aggs query, '
                                   '{doc_query_default} for only-docs query).'
                                   .format(doc_query_default=N_DOC_DEFAULT))
    query_parser.add_argument('-t', '--term', nargs='*',
                              metavar='<term name> <term value>',
                              help='term query.')
    # aggs
    query_parser.add_argument('-N', '--agg-number', type=int, default=20,
                              help='size of agg y(default: %(default)s).')
    query_parser.add_argument('-x', '--agg-x', nargs='*',
                              help='x of the aggregation(built-in: %s, '
                                   'or custom: `<agg_func> <field>`).' % builtin_x)
    query_parser.add_argument('-y', '--agg-y', '--by', nargs='*',
                              help='y of the aggregation(built-in: %s, '
                                   'or custom: `<field>`).' % builtin_y)
    query_parser.add_argument('-m', '--agg-multi-y', '--by-multi', nargs='*',
                              help='y of the aggregation, multi field')
    query_parser.add_argument('-S', '--agg-script-y', '--by-script',
                              help='y of the aggregation, script')
    query_parser.add_argument('--min-value', '--min',
                              help='filter result entires with the min value')
    query_parser.set_defaults(func=query)

    args = parser.parse_args()
    args.user = args.user or get_os_username()

    if args.func == query:
        if args.term and len(args.term) % 2:
            parser.error('-t/--term should accept name/value pairs.')
        if not args.agg_x and (args.agg_y or args.agg_multi_y or args.agg_script_y):
            parser.error('-x/--agg-x should be provided.')
        if args.start is None:
            parser.error('-s/--start invalid, must be negative integer, or datetime string')

    if args.func == analyze:
        if args.term and len(args.term) % 2:
            parser.error('-t/--term should accept name/value pairs.')
        if not args.agg_y and not args.agg_multi_y and not args.agg_script_y:
            parser.error('aggs y should be provided.')
        if args.start_abnormal is None or args.start_normal is None:
            parser.error('-a, -b invalid, must be negative integer, or datetime string')

    levels = [level for v, level in VERBOSE_LEVEL if args.verbose >= v]
    level = min(levels)
    logging.getLogger("requests").setLevel(level)
    logging.getLogger("elasticsearch").setLevel(level)
    logging.basicConfig(level=level,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    sys.exit(args.func(args))



# coding: utf-8

from __future__ import print_function

import time
from datetime import datetime

from elasticsearch import Elasticsearch, __version__ as es_client_version


class ESQuery(object):
    def __init__(self, es_hosts, index_prefix, timestamp_field, index_time_format='%Y.%m.%d', es_user='', es_passwd=''):
        self.client = Elasticsearch(hosts=es_hosts, http_auth=(es_user, es_passwd)) if es_user and es_passwd else Elasticsearch(hosts=es_hosts)
        self.index_prefix = index_prefix
        self.timestamp_field = timestamp_field
        self.index_time_format = index_time_format

    def get_mapping(self, fields=None):
        """Retrieve mapping definition of index or specific field.
        http://www.elastic.co/guide/en/elasticsearch/reference/current/indices-get-mapping.html"""
        indexes = self.compute_indexes(*self.compute_start_end_timestamp(0, 0))
        if fields:
            res = self.client.indices.get_field_mapping(index=indexes,
                                                        fields=fields)
        else:
            res = self.client.indices.get_mapping(index=indexes)
        return res

    def compute_start_end_timestamp(self, start, duration):
        if isinstance(start, int):
            now_timestamp = int(time.time())
            start_timestamp = now_timestamp + start * 60
        else:
            start_timestamp = int(time.mktime(start.timetuple()))
        end_timestamp = start_timestamp + duration * 60
        return start_timestamp, end_timestamp

    def compute_time_range(self, start_timestamp, end_timestamp):
        time_range = {self.timestamp_field: {'gte': start_timestamp,
                                             'lte': end_timestamp,
                                             'format': "epoch_second"}}
        return time_range

    def compute_indexes(self, start_timestamp, end_timestamp):
        start_utc_datetime = datetime.utcfromtimestamp(start_timestamp)
        end_utc_datetime = datetime.utcfromtimestamp(end_timestamp)
        ordinal_range = range(start_utc_datetime.toordinal(), end_utc_datetime.toordinal() + 1)
        indexes = ','.join('%s%s' % (self.index_prefix,
                                     datetime.fromordinal(i).strftime(self.index_time_format))
                           for i in ordinal_range)
        return indexes

    def make_body(self, time_range=None, query_string=None, term_dict=None,
                  aggregations=None, sort=None,
                  **kw):
        must_list = [{"range": time_range}]
        if term_dict:
            must_list += [{"term": {k: v}} for k, v in term_dict.items()]
        body = {
            "query": {
                "bool": {
                    "must": {
                        "query_string": query_string,
                    },
                    "filter": {
                        "bool": {
                            "must": must_list,
                            # "must_not": [],
                        }
                    }
                }
            },
            "aggs": aggregations,
            "sort": sort,
        }
        return body

    @staticmethod
    def _filter_None(obj):
        if obj is None or isinstance(obj, (int, str)):
            return obj
        if isinstance(obj, list):
            return [ESQuery._filter_None(i) for i in obj if ESQuery._filter_None(i)]
        return {k: ESQuery._filter_None(v) for k, v in obj.items() if ESQuery._filter_None(v)}

    def query(self, lucene_query_string="*", term_dict=None, aggregations=None, sort=None,
              start=-15, duration=15, size=0):
        start_timestamp, end_timestamp = self.compute_start_end_timestamp(start, duration)
        time_range = self.compute_time_range(start_timestamp, end_timestamp)
        indexes = self.compute_indexes(start_timestamp, end_timestamp)

        if lucene_query_string and lucene_query_string != '*':
            query_string = dict(query=lucene_query_string,
                                analyze_wildcard=True)
        else:
            query_string = None

        body = self.make_body(time_range=time_range, query_string=query_string,
                              term_dict=term_dict,
                              aggregations=aggregations, sort=sort)
        body = ESQuery._filter_None(body)

        params = dict(index=indexes,
                      size=size,
                      body=body)

        res = self.client.search(**params)
        return res


if __name__ == '__main__':
    body = {
        "query": {
            "bool": {
                "must": {
                    "query_string": 'test',
                },
                "filter": {
                    "bool": {
                        "must": [
                            {"range": 'yes'},
                            {"term": None},
                        ],
                        "must_not": [],
                    }
                }
            }
        },
        "aggs": None,
    }
    print(ESQuery._filter_None(body))



# coding: utf-8

import os
import re
import sys
import json
import pkgutil
import logging
import argparse
import importlib
from html.parser import HTMLParser

import yaml
import inflection

from sa_tools_core.consts import TENCENT_DEFAULT_PARAMS, TENCENT_DEFAULT_REGIN

from sa_tools_core.utils import get_config

logger = logging.getLogger(__name__)

tencent_config_func = lambda: get_config('tencent')  # NOQA

RE_PARAM = re.compile(r'^:(param|type)\ (\w+):\ (.*)$')
COMMON_PARAM_TYPES = {
    'str': str,
    'int': int,
    'bool': bool,
    'int non-negative': int,
}


class KVParamType:
    def __init__(self, kname, vname, multi_value=False):
        self.kname = kname
        self.vname = vname
        self.multi_value = multi_value

    def parse_value(self, arg):
        k, v = arg.split('=')
        return {self.kname: k, self.vname: v.split(',') if self.multi_value else v}

    @property
    def help_message(self):
        if self.multi_value:
            return f'\npattern: {self.kname}={self.vname},{self.vname}...'
        else:
            return f'\npattern: {self.kname}={self.vname}'


SPECIAL_PARAM_TYPES = {
    'Filter': KVParamType('Name', 'Values', True),
    'Tag': KVParamType('TagKey', 'TagValues', True),
    'DeviceAlias': KVParamType('InstanceId', 'Alias', False),
}

# TODO: deal with list of internal type
IGNORE_PARAM_TYPES = [
    'DataDisk',
    'TagSpecification',
]


def simplify_action(action):
    action = action.replace('Get', '') \
        .replace('Describe', '') \
        .replace('Description', '') \
        .replace('UnBind', 'Unbind') \
        .replace('LoadBalancer', 'Lb') \
        .replace('Ex', '') \
        .replace('Query', '')
    return inflection.underscore(action)


class HTMLFilter(HTMLParser):
    text = ""

    def handle_starttag(self, tag, attrs):
        if tag == 'li':
            self.text += '\n-> '
        elif tag == 'br':
            self.text += '\n'
        elif tag == 'p':
            self.text += '\n'

    def handle_data(self, data):
        self.text += data


def cleanup_help_message(msg):
    f = HTMLFilter()
    f.feed(msg)
    return f.text.replace(r"%", r"%%").replace('\n\n', '\n')


class ServiceClient:
    def __init__(self, service, version, cls_):
        self.service = service
        self.version = version
        self.cls = cls_


class ParamInfo:
    def __init__(self, name):
        self.name = name
        self.type = None
        self.desc = None
        self.multi = False


def add_output_format_args(parser):
    parser.add_argument(
        '-f',
        '--format',
        default='plain',
        choices=('plain', 'json', 'yaml', 'raw'),
        help='output format, (default: %(default)s)',
    )
    parser.add_argument(
        '-s',
        '--sep',
        default=', ',
        help='separator, only apply in plain output format (default: %(default)s)',
    )
    parser.add_argument('-a', '--attrs', nargs='*', help='the attrs should only output')
    parser.add_argument('-e', '--excludes', nargs='*', help='the extra-attrs that should not output.')


def extract_params(doc_str, models):
    '''
    extract params from doc string
    '''
    params = dict()
    last_pname = ''
    for line in doc_str.splitlines():
        # NOTE:(everpcpc) for multiline doc
        if (not line.startswith(" ")) and last_pname:
            params[last_pname].desc += '\n' + line

        m = RE_PARAM.match(line.strip())
        if not m:
            continue
        last_pname = pname = inflection.underscore(m.group(2))
        pinfo = params.pop(pname, None) or ParamInfo(m.group(2))
        if m.group(1) == 'param':
            pinfo.desc = m.group(3)
        elif m.group(1) == 'type':
            t = m.group(3)
            if t.startswith('list of '):
                pinfo.multi = True
                t = t[len('list of '):]

            if t in COMMON_PARAM_TYPES:
                pinfo.type = COMMON_PARAM_TYPES[t]
            elif t in SPECIAL_PARAM_TYPES.keys():
                pinfo.type = getattr(models, t)
            # NOTE: deal with internal class
            elif t in IGNORE_PARAM_TYPES:
                continue
            elif t.startswith(':class:'):
                _mod, _t = t[len(':class:'):].strip('`').rsplit('.', 1)
                if _mod != models.__name__:
                    raise Exception(f'param type in other models: {line.strip()}')
                subparams = extract_params(getattr(models, _t).__init__.__doc__, models)
                pinfo.type = subparams
            else:
                raise Exception(f'unkown param type {t} => {line.strip()}')
        if pinfo:
            params[pname] = pinfo

    return params


def param2parser(parser, param, info):
    '''
    translate param into parser
    '''
    if isinstance(info.type, dict):
        for sp, si in info.type.items():
            param2parser(parser, f'{param}_{sp}', si)
        return

    kw = {'help': cleanup_help_message(info.desc)}
    tname = info.type.__name__
    if info.multi:
        kw['nargs'] = '*'
    if tname in COMMON_PARAM_TYPES.keys():
        if tname == 'bool':
            default_value = False
            if param in TENCENT_DEFAULT_PARAMS.keys():
                default_value = TENCENT_DEFAULT_PARAMS[param]
            elif '默认取值：TRUE' in info.desc:
                default_value = True

            if default_value:
                kw['action'] = 'store_false'
            else:
                kw['action'] = 'store_true'
            kw['help'] += ' (default: %(default)s)'

        else:
            if tname == 'int':
                kw['type'] = int

            if param in TENCENT_DEFAULT_PARAMS.keys():
                kw['default'] = TENCENT_DEFAULT_PARAMS[param]
                kw['help'] += ' (default: %(default)s)'
    # param type in sdk
    elif tname in SPECIAL_PARAM_TYPES.keys():
        kw['help'] += SPECIAL_PARAM_TYPES[tname].help_message
    else:
        # NOTE:(everpcpc) if raised, add support for it
        raise Exception(f'param: {info.name} => {info.type} not yet supported')
    parser.add_argument(f'--{inflection.dasherize(param)}', **kw)


def arg2param(arg, param, info):
    tname = info.type.__name__
    if tname in COMMON_PARAM_TYPES.keys():
        return arg
    elif tname in SPECIAL_PARAM_TYPES.keys():
        if info.multi:
            return [SPECIAL_PARAM_TYPES[tname].parse_value(a) for a in arg]
        return SPECIAL_PARAM_TYPES[tname].parse_value(arg)
    else:
        raise Exception(f'param {param} not supported: {info}')


def args2params(args, params, prefix=''):
    '''
    translate argument into request param
    '''
    req_params = dict()
    for param, info in params.items():
        if isinstance(info.type, dict):
            req_params[info.name] = args2params(args, info.type, prefix=f'{prefix}{param}_')
            continue

        arg = getattr(args, f'{prefix}{param}', None)
        if arg is not None:
            req_params[info.name] = arg2param(arg, param, info)

    return req_params


def simplify_output(output, oformat, sep=', ', attrs=(), excludes=()):
    data = json.loads(output)

    def _simplify(data):
        if isinstance(data, list):
            data = [_simplify(i) for i in data]
            if oformat == 'plain':
                data = '\n'.join(sorted(i for i in data))
        elif isinstance(data, dict):
            is_leaf = False
            # NOTE: rough leaf determination
            if not any(isinstance(v, (dict,)) for k, v in data.items()):
                is_leaf = True
            if any(isinstance(i, (dict,)) for k, v in data.items() if isinstance(v, list) for i in v):
                is_leaf = False

            if is_leaf:
                if attrs:
                    data = {k: v for k, v in data.items() if k in attrs}
                else:
                    data = {k: v for k, v in data.items() if k not in excludes}
                if oformat == 'plain':
                    data = sep.join([str(v) for k, v in data.items()])
            else:
                data = {k: _simplify(v) for k, v in data.items()} or data
                if oformat == 'plain':
                    data = '\n'.join([
                        ('%s' if idx % 2 else '>> %s') % i for p in data.items() for idx, i in enumerate(p)
                    ])
        return data

    return _simplify(data)


def _execute(req_cls, cli_cls, action, params):
    from tencentcloud.common import credential
    from tencentcloud.common.profile.http_profile import HttpProfile
    from tencentcloud.common.profile.client_profile import ClientProfile

    _id, _key = tencent_config_func().split(':')
    cred = credential.Credential(_id, _key)
    profile = ClientProfile(httpProfile=HttpProfile())
    client = cli_cls(cred, TENCENT_DEFAULT_REGIN, profile)

    req = req_cls()
    req.from_json_string(json.dumps(params))
    return getattr(client, action)(req)


def execute_action(client, action, argv):
    models = importlib.import_module(f'tencentcloud.{client.service}.{client.version}.models')
    request_cls = getattr(models, f'{action}Request')
    # NOTE:(everpcpc) extract params with action request document
    params = extract_params(request_cls.__init__.__doc__, models)

    parser = argparse.ArgumentParser(
        f'{os.path.basename(sys.argv[0])} {client.service} {action}',
        formatter_class=argparse.RawTextHelpFormatter,
    )
    add_output_format_args(parser)

    for param, info in params.items():
        param2parser(parser, param, info)

    args = parser.parse_args(argv)

    req_params = args2params(args, params)

    ret = _execute(request_cls, client.cls, action, req_params)

    if args.format == 'raw':
        print(ret)
        return

    simplified = simplify_output(
        ret.to_json_string(),
        args.format,
        args.sep,
        args.attrs or (),
        args.excludes or (),
    )
    if args.format == 'plain':
        print(simplified)
    elif args.format == 'json':
        print(json.dumps(simplified, ensure_ascii=False, indent=2))
    elif args.format == 'yaml':
        print(yaml.dump(simplified, allow_unicode=True, indent=2))


def find_service_version(service):
    service_mod = importlib.import_module(f'tencentcloud.{service}')
    for m in pkgutil.iter_modules(service_mod.__path__):
        if m.ispkg:
            return m.name


def execute_service(service, argv):
    parser = argparse.ArgumentParser(
        f'{os.path.basename(sys.argv[0])} {service}',
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(help='Subcommands', dest='action')
    subparsers.required = True

    version = find_service_version(service)
    if not version:
        raise Exception(f"module not found for {service}")

    mapping = dict()
    client_mod = importlib.import_module(f'tencentcloud.{service}.{version}.{service}_client')
    client_cls = getattr(client_mod, f'{service.capitalize()}Client')

    for method in dir(client_cls):
        if not method[0].isupper():
            continue
        simple_action = simplify_action(method)
        mapping[simple_action] = method
        desc = getattr(client_cls, method).__doc__.split("\n\n")[0].strip()
        subparsers.add_parser(simple_action, help=cleanup_help_message(desc))

    args = parser.parse_args(argv[:1])

    return execute_action(ServiceClient(service, version, client_cls), mapping[args.action], argv[1:])


def main():
    """
    e.g.

    sa-tc bm devices
    sa-tc bm devices -f yaml
    sa-tc bm devices --alias host
    """

    parser = argparse.ArgumentParser(epilog=main.__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-v', '--verbose', action='count', default=0, help='verbose.')

    subparsers = parser.add_subparsers(help='Services', dest='service')
    subparsers.required = True

    import tencentcloud
    for service in pkgutil.iter_modules(tencentcloud.__path__):
        if service.ispkg:
            subparsers.add_parser(service.name, help=f'service => {service.name}')

    args = parser.parse_args(sys.argv[1:2])

    level = logging.WARNING - args.verbose * 10
    logging.basicConfig(level=level, format='%(asctime)s %(name)s %(levelname)s %(message)s')

    # NOTE:(everpcpc) return code not parsed now
    sys.exit(execute_service(args.service, sys.argv[2:]))



# coding: utf-8

import logging

from icinga2_api.api import Api

from sa_tools_core.consts import ICINGA_CACERT
from sa_tools_core.utils import get_config

logger = logging.getLogger(__name__)

icinga_api = None


class IcingaClusterConfig(object):
    """you need to inhert IcingaClusterConfig and impl your own config class
    and specify your class path in sa_tools_core.consts.ICINGA_CLUSTER_CONFIG_CLASS
    """
    @classmethod
    def get_ack_link(cls, env):
        return ''

    @classmethod
    def get_reboot_host_link(cls, env):
        return ''

    @classmethod
    def get_icinga_link(cls, env):
        return ''

    @classmethod
    def get_icinga_hosts(cls):
        return []


class IcingaApi(object):
    def __init__(self, icinga_hosts, icinga_auth, icinga_cacert):
        self.api = Api(icinga_hosts,
                       icinga_auth,
                       icinga_cacert)
        self._filters = []

    def _filter(self, name, value, regex=False, wildcard=False, negation=False, operator='=='):
        if not value:
            return
        value = '"%s"' % value
        filter_ = ''
        if regex:
            filter_ = 'regex(%s, %s)' % (value, name)
        elif wildcard:
            filter_ = 'match(%s, %s)' % (value, name)
        else:
            filter_ = '%s %s %s' % (name, operator, value)
        if negation:
            filter_ = '! (%s)' % filter_
        self._filters.append(filter_)
        return filter_

    @property
    def filter(self):
        return ' && '.join(self._filters)

    def clear_filters(self):
        self._filters = []

    def acknowledge(self, host, service=None, author='anonymous', comment='ack by sa-icinga',
                    remove=False, notify=False):
        type_ = 'Service' if service else 'Host'
        self.clear_filters()
        if host and '*' in host:
            self._filter('host.name', host, wildcard=True)
        else:
            self._filter('host.name', host)
        if service and '*' in service:
            self._filter('service.name', service, wildcard=True)
        else:
            self._filter('service.name', service)

        params = dict(filter=self.filter,
                      type=type_)
        if not remove:
            action = 'acknowledge-problem'
            params['author'] = author
            params['comment'] = comment
        else:
            action = 'remove-acknowledgement'
        if notify:
            params['notify'] = notify

        res = (self.api
               .actions.url(action)
               .post(**params))
        results = res.get('results')
        if results and int(results[0].get('code')) == 200:
            return True, res
        else:
            return False, res


def get_icinga_api(icinga_cluster_config):
    global icinga_api
    if icinga_api is None:
        api_user, passwd = get_config('icinga').split(':')
        ICINGA_AUTH = (api_user, passwd)
        icinga_api = IcingaApi(icinga_cluster_config.get_icinga_hosts(),
                               ICINGA_AUTH,
                               ICINGA_CACERT)
    return icinga_api



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
https://github.com/tclh123/process

# Usage:

from process import process

git = process.git
git.status()

ls = process.bake('ls')
ls('-al')
ls('-l', '-a')
ls('-l', a=True)
ls.call('-al')
process.ls.call('-al')


## Keyword arguments like http://amoffat.github.io/sh/#keyword-arguments

# resolves to "curl http://duckduckgo.com/ -o page.html --silent"
curl("http://duckduckgo.com/", o="page.html", silent=True)

# or if you prefer not to use keyword arguments, this does the same thing:
curl("http://duckduckgo.com/", "-o", "page.html", "--silent")

# resolves to "adduser amoffat --system --shell=/bin/bash --no-create-home"
adduser("amoffat", system=True, shell="/bin/bash", no_create_home=True)

# or
adduser("amoffat", "--system", "--shell", "/bin/bash", "--no-create-home")
"""

import six
import shlex
import logging
import subprocess

logger = logging.getLogger(__name__)


def _call(cmd, env=None, nonblock=False, shell=False):
    '''if nonblock, return the process itself, otherwise return a result dict'''
    fullcmd = cmd if shell else ' '.join(cmd)
    logger.debug("Running process %s" % fullcmd)
    kw = dict(env=env) if env else {}
    try:
        process = subprocess.Popen(cmd, shell=shell, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   universal_newlines=True, **kw)
    except (OSError, TypeError) as err:
        logger.error("Error occurrred when calling %s" % fullcmd)
        raise err
    if nonblock:
        return process
    out, err = process.communicate()
    out = str(out)
    err = str(err)

    result = {}
    result['returncode'] = process.returncode
    result['stdout'] = out
    result['stderr'] = err
    result['fullcmd'] = fullcmd
    return result


# TODO: add tests or doctests
class Process(object):
    def __init__(self, cmds=None):
        self.cmds = cmds or []

    def __getattr__(self, name):
        name = name.replace('_', '-')  # e.g. format_patch -> format-patch
        return self.bake(name)

    def __call__(self, *a, **kw):
        proc = self.bake()
        env = kw.pop('env', {})
        proc._parse_args(*a, **kw)
        return proc.call(env=env)

    def _parse_args(self, *a, **kw):
        cmds = []
        for p in a:
            if not isinstance(p, six.string_types):
                raise KeyError
            cmds.append(p)

        for k, v in kw.items():
            if len(k) == 1:
                k = '-' + k
            else:
                k = '--' + k
            if '_' in k:  # e.g. --no_ff -> --no-ff
                k = k.replace('_', '-')
            if not v:  # v in (None, '', False)
                continue
            elif isinstance(v, bool):  # v is True
                cmds.append(k)
            elif isinstance(v, six.string_types):
                cmds.append(k)
                cmds.append(v)
            else:
                raise KeyError
        self.cmds += cmds

    def bake(self, *a, **kw):
        cmds = list(self.cmds)
        proc = Process(cmds)
        proc._parse_args(*a, **kw)
        return proc

    def call(self, cmdstr='', env=None, nonblock=False, shell=False):
        if not shell:
            extra_cmds = shlex.split(cmdstr)
            cmd = self.cmds + extra_cmds
        else:
            cmd = ' '.join(self.cmds)
            if cmd:
                cmd += ' ' + cmdstr
            else:
                cmd = cmdstr
        return _call(cmd, env=env, nonblock=nonblock, shell=shell)


process = Process()



import requests
import base64
import json
import logging
import datetime
from pathlib import Path, PurePath

from sa_tools_core.consts import GITHUB_API_ENTRYPOINT
from sa_tools_core.utils import get_config

logger = logging.getLogger(__name__)


def github_secret_func():
    return get_config('github')


class GithubRepo:
    def __init__(self, org, repo, entrypoint=None,
                 user_name=None, personal_token=None, secret_func=None,
                 author=None, skip_ssl=False):
        self.org = org
        self.repo = repo
        self.author = author
        self.entrypoint = entrypoint
        self.base_commit = None
        self.base_tree = None
        self.head_commit = None
        self.head_tree = None
        self.secret_func = secret_func
        self.session = requests.Session()
        if user_name and personal_token:
            self.user_name = user_name
            self.personal_token = personal_token
        else:
            (self.user_name, self.personal_token) = self.get_user_token_pair()
        self.session.auth = (self.user_name, self.personal_token)
        self.skip_ssl = skip_ssl

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if self.session:
            self.session.close()

    def get_user_token_pair(self):
        if self.secret_func:
            secret = self.secret_func()
            divided_secret = secret.split(':', 1)
            if len(divided_secret) != 2:
                raise ValueError("Invalid secret, suppose to be able to be split into 2 strings")
            return divided_secret
        raise ValueError("Please define either username & token or secret_func.")

    def make_request(self, method, api_path, **kwargs):
        r = self.session.request(method, f"{self.entrypoint}{api_path}", verify=(not self.skip_ssl), **kwargs)
        try:
            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            logger.error(r.text)
            raise e
        if r.status_code == 204:
            return None
        return r.json()

    def get_file(self, path, reference=None):
        """get content
        GET /repos/:owner/:repo/contents/:path
        :: params
        reference: str branch, tag or commit id, default: default branch of repo
        path: str relative path
        :: return
        dict
        """
        try:
            return self.make_request('GET', f"/repos/{self.org}/{self.repo}/contents/{path}",
                                 params={'ref': reference or 'master'})
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return None
            raise e

    def get_reference(self, reference):
        return self.make_request('GET', f"/repos/{self.org}/{self.repo}/git/ref/heads/{reference}")

    def get_commit(self, commit_sha):
        return self.make_request('GET', f"/repos/{self.org}/{self.repo}/git/commits/{commit_sha}")

    def update_a_file(self, path, content, message, sha=None):
        return self.make_request('PUT', f"/repos/{self.org}/{self.repo}/contents/{path}",
                                 data=json.dumps({
                                     'message': message,
                                     'content': base64.b64encode(content).decode(),
                                     'sha': sha
                                 }))

    def upload_one_file(self, content):
        """
        POST /repos/:owner/:repo/git/blobs
        :: input
        {
            "content": "Content of the blob",
            "encoding": "utf-8"
        }
        :: return
        {
            "url": "https://api.github.com/repos/octocat/example/git/blobs/3a0f86fb8db8eea7ccbb9a95f325ddbedfb25e15",
            "sha": "3a0f86fb8db8eea7ccbb9a95f325ddbedfb25e15"
        }

        """
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/git/blobs",
                                 data=json.dumps({
                                     'content': base64.b64encode(content).decode(),
                                     'encoding': 'base64'
                                 }))

    def create_tree(self, base_tree, files):
        """ equivalent to  git add
        """
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/git/trees",
                                 data=json.dumps({
                                     'base_tree': base_tree,
                                     'tree': files
                                 }))

    def create_commit(self, base, tree, message):
        """ equivalent to git commit
        """
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/git/commits",
                                 data=json.dumps({
                                     'message': message,
                                     'tree': tree,
                                     'parents': [base]
                                 }))

    def create_reference(self, reference, commit_sha):
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/git/refs",
                                 data=json.dumps({
                                     'ref': f"refs/heads/{reference}",
                                     'sha': commit_sha
                                 }))

    def update_reference(self, reference, commit_sha):
        """ equivalent to git push
        PATCH /repos/:owner/:repo/git/refs/:ref
        """
        return self.make_request('PATCH', f"/repos/{self.org}/{self.repo}/git/refs/heads/{reference}",
                                 data=json.dumps({
                                     "sha": commit_sha,
                                     'force': False
                                 }))

    def create_pull_request(self, title, head, base, body='', maintainer_can_modify=True, draft=False, ):
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/pulls",
                                 data=json.dumps({
                                     'title': title,
                                     'head': head,
                                     'base': base,
                                     'body': body,
                                     'maintainer_can_modify': maintainer_can_modify,
                                     'draft': draft
                                 }))

    def merge_pull_request(self, pull_number, merge_method='merge'):
        """merge method: `rebase`, `squash`, `merge`"""
        return self.make_request('PUT', f"/repos/{self.org}/{self.repo}/pulls/{pull_number}/merge",
                                 data=json.dumps({
                                     'merge_method': merge_method
                                 }))

    def has_collaborator(self, username):
        """if present, return True, if not, return False"""
        r = self.session.request("GET", f"{self.entrypoint}/repos/{self.org}/{self.repo}/collaborators/{username}",
                                 verify=(not self.skip_ssl))
        if r.status_code == 204:
            return True
        elif r.status_code == 404:
            return False
        else:
            r.raise_for_status()

    def add_collaborator(self, username, permission=None):
        """permissions: `pull`, `push`, `admin`, `maintain`, `triage` 
        return None on default, raise exception if anything wrong
        """
        if permission is None:
            permission = 'push'
        self.make_request('PUT', f"/repos/{self.org}/{self.repo}/collaborators/{username}",
                                 data=json.dumps({
                                     'permission': permission
                                 }))

    def remove_collaborator(self, username):
        self.make_request('DELETE', f"/repos/{self.org}/{self.repo}/collaborators/{username}")

    def create_commit_comment(self, commit_sha, comment_content, path=None, position=None):
        comment = {'body': comment_content}
        if path:
            comment['path'] = path
        if position:
            comment['position'] = position
        return self.make_request('POST', f"/repos/{self.org}/{self.repo}/commits/{commit_sha}/comments",
                                 data=json.dumps(comment))

    def list_commit_comments(self, commit_sha):
        return self.make_request('GET', f"/repos/{self.org}/{self.repo}/commits/{commit_sha}/comments")

    # high level api starts here
    def add(self, files, base_reference):
        """add files to the tree, generate a tree , store the tree_sha to self.head_tree"""
        files_sha = []
        # upload
        for path, content in files.items():
            base_file = self.get_file(path, reference=base_reference)
            if base_file and base64.b64decode(base_file['content'].encode()) == content:
                logger.info(f'{path} unchange, ignored')
                continue
            upload_result = self.upload_one_file(content)
            files_sha += [{
                'path': path,
                'sha': upload_result['sha'],
                'type': 'blob',
                'mode': '100644'
            }]
        if len(files_sha) == 0:
            logger.info("changelist empty, ignored")
            return

        # add to index, need tree base sha
        base_branch = self.get_reference(base_reference)
        self.base_commit = self.get_commit(base_branch['object']['sha'])
        self.head_tree = self.create_tree(self.base_commit['tree']['sha'], files_sha)
        return self.head_tree

    def commit(self, message):
        self.head_commit = self.create_commit(self.base_commit['sha'], self.head_tree['sha'], message)
        logger.info('commit create: %s', self.head_commit['url'])

    def push(self, reference):
        try:
            self.update_reference(reference, self.head_commit['sha'])
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 422:
                # error 422 , reference not exist, create reference
                self.create_reference(reference, self.head_commit['sha'])
            else:
                raise e

    def download(self, remote_path, local_path, reference='master'):
        """download a folder recursively, only file and directory supported, """
        file_or_dir = self.get_file(remote_path)
        if isinstance(file_or_dir, dict):
            # file, submodule, link, only file is supported now
            if file_or_dir['type'] != 'file':
                logger.warning('Only file and directory are supported, submodule and symlink are ignored.')
                logger.warning(f"{file_or_dir['type']} ignored : {file_or_dir['path']}")
                return
            with open(local_path, 'wb') as fp:
                fp.write(base64.b64decode(
                    file_or_dir['content'].encode()))
        else:
            # dir
            Path(local_path).mkdir(parents=True, exist_ok=True)
            for f in file_or_dir:
                self.download(f['path'], PurePath(local_path).joinpath(f['name']), reference=reference)

    def update_files(self, reference, files, message):
        """update reference
        reference: str branch, tag
        files: dict {file_path, filecontent}
            file_path -> str
            filecontent -> bytes
            example: {'example.md', b'Hello world', 'libs/example2.md', b'Hello world2'}
        possible error:
        Web failed, timeout, 404, 403.
        None fast forward.
        No change.
        """
        if len(files.keys()) == 1:
            # simple update file
            path = list(files.keys())[0]
            content = files[path]
            base_file = self.get_file(path, reference=reference)
            if not base_file:
                self.update_a_file(path, content, message)
            elif base64.b64decode(base_file['content'].encode()) == content:
                logger.info(f'{path} unchange, ignored')
            else:
                self.update_a_file(path, content, message, sha=base_file['sha'])
            return

        add_result = self.add(files, reference)
        if add_result:
            # do not continue if add not successful
            self.commit(message)
            self.push(reference)

    def submit_pr(self, base_branch, files, title,
                  body='', commit_message='', new_branch_name='',
                  auto_merge=False, merge_method='merge'):
        if not commit_message:
            commit_message = "Update {}".format(', '.join(files.keys()))
        if not new_branch_name:
            new_branch_name = "patch-{}".format(datetime.datetime.now().strftime('%m-%d.%H.%M'))
        add_result = self.add(files, base_branch)
        if not add_result:
            logger.info('Empty changelist, ignored')
            return
        # do not continue if add not successful
        self.commit(commit_message)
        self.push(new_branch_name)
        created_pr = self.create_pull_request(title, new_branch_name, base_branch, body=body)
        logger.info(f"PR create, url: {created_pr['html_url']}")
        if auto_merge:
            self.merge_pull_request(created_pr['number'], merge_method=merge_method)
            logger.info('PR was successfully merged')
        return created_pr
    
    def ensure_collaborator(self, username, permission=None):
        """permissions: `pull`, `push`, `admin`, `maintain`, `triage` 
        return None on default, raise exception if anything wrong
        """
        if self.has_collaborator(username):
            self.remove_collaborator(username)
        self.add_collaborator(username, permission=permission)


def commit_github(org, repo, branch, files, message, retry=2):
    """update files of a repo on a branch head.
    ::params:
        org: str
        repo: str
        branch: str
        files: dict {file_path, filecontent}
            file_path -> str
            filecontent -> bytes
            example: {'example.md', b'Hello world', 'libs/example2.md', b'Hello world2'}
    """
    return_value = -1
    with GithubRepo(org, repo, entrypoint=GITHUB_API_ENTRYPOINT, secret_func=github_secret_func) as gh:
        for _ in range(retry + 1):
            try:
                gh.update_files(branch, files, message)
                return_value = 0
                break
            except Exception as e:
                logger.warning(e)
                logger.warning('Request failed , retrying')
    return return_value


def submit_pr(org, repo, *args, **kwargs):
    with GithubRepo(org, repo, entrypoint=GITHUB_API_ENTRYPOINT, secret_func=github_secret_func) as gh:
        return gh.submit_pr(*args, **kwargs)



import os
import subprocess
import tempfile


class Editor(object):

    def __init__(self, editor=None, env=None, require_save=True,
                 extension='.txt'):
        self.editor = editor
        self.env = env
        self.require_save = require_save
        self.extension = extension

    def get_editor(self):
        if self.editor is not None:
            return self.editor
        for key in 'VISUAL', 'EDITOR':
            rv = os.environ.get(key)
            if rv:
                return rv
        for editor in 'vim', 'nano':
            if os.system('which %s >/dev/null 2>&1' % editor) == 0:
                return editor
        return 'vi'

    def edit_file(self, filename):
        editor = self.get_editor()
        if self.env:
            environ = os.environ.copy()
            environ.update(self.env)
        else:
            environ = None
        try:
            c = subprocess.Popen('%s "%s"' % (editor, filename),
                                 env=environ, shell=True)
            exit_code = c.wait()
            if exit_code != 0:
                print('%s: Editing failed!' % editor)
        except OSError as e:
            print('%s: Editing failed: %s' % (editor, e))

    def edit(self, text=None):
        text = text or ''
        if text and not text.endswith('\n'):
            text += '\n'

        fd, name = tempfile.mkstemp(prefix='editor-', suffix=self.extension)
        try:
            encoding = 'utf-8'
            text = text.encode(encoding)

            f = os.fdopen(fd, 'wb')
            f.write(text)
            f.close()
            timestamp = os.path.getmtime(name)

            self.edit_file(name)

            if self.require_save \
               and os.path.getmtime(name) == timestamp:
                return None

            f = open(name, 'rb')
            try:
                rv = f.read()
            finally:
                f.close()
            return rv.decode('utf-8-sig').replace('\r\n', '\n')
        finally:
            os.unlink(name)



# coding: utf-8

import logging

logger = logging.getLogger(__name__)


def send_sms(num, content):
    # implement this yourself
    # TODO: use plugin mode
    logger.info('Success send sms to %s', num)



# coding: utf-8

from __future__ import absolute_import

import json
import logging

import pylark

from sa_tools_core.utils import get_config
from sa_tools_core.consts import DEFAULT_LARK_TENANT

logger = logging.getLogger(__name__)

APP_ID = "aaaaaaaaaaaaaaaaaa"
APP_SECRET = "bbbbbbbbbbbbbbbbbbbbbb"


class LarkApp(object):
    tenants = {}

    @staticmethod
    def load_from_configs(configs: dict):
        app = LarkApp()
        for name, config in configs.items():
            lark = pylark.Lark(
                app_id=config["app_id"],
                app_secret=config["app_secret"]
            )
            app.tenants[name] = lark
        return app

    def get_lark(self, company="douban"):
        lark = self.tenants.get(company)
        if lark:
            return lark
        logger.error(f"no tenant named {company}, please check your code and configs")


def send_message(addrs, content, **kwargs):
    secret = get_config("lark-multi-tenant")
    # config template:
    # {
    #     "your_company": {
    #        "app_id": "1",
    #        "app_secret": "1",
    #        "encrypt_key": "1",
    #        "verification_token": "1"
    #    }
    # }
    lark_bundle = LarkApp.load_from_configs(configs=json.loads(secret))
    company = kwargs.get("company", DEFAULT_LARK_TENANT)
    lark = lark_bundle.get_lark(company=company)
    for addr in addrs:
        res, response = lark.message.send_raw_message(
            pylark.SendRawMessageReq(
                receive_id_type="email",
                receive_id=addr,
                content=json.dumps({"text": content}),
                msg_type="text",
            )
        )



# coding: utf-8
# TODO:
#   1. LDAP backend auth
#   2. cli & python auth credentials

import sys
import logging


logger = logging.getLogger(__name__)


def require_user(func):
    def _require_user(args, *a, **kw):
        if auth_user(args.user):
            args.authed_user = args.user
            return func(args, *a, **kw)
        else:
            sys.exit(1)
    return _require_user


def require_sa(func):
    @require_user
    def _require_sa(args, *a, **kw):
        # if args.authed_user in SAs:
        if True:
            return func(args, *a, **kw)
        else:
            logger.error('Needs SA permission.')
            sys.exit(1)
    return require_user(_require_sa)


def auth_user(user, password=None):
    return True



# coding: utf-8

from __future__ import print_function

import os
import re
import json
import fcntl
import logging
import datetime
import argparse

import humanize

from sa_tools_core.libs.permission import require_sa
from sa_tools_core.libs.process import process
from sa_tools_core.libs.ncdu import top_huge_dirs_from_ncdu
from sa_tools_core.utils import get_os_username, to_str, prompt_input
from sa_tools_core.consts import NCDU_EXPORT_DATA_PATH, NCDU_JOB_LOCK_PREFIX

logger = logging.getLogger(__name__)
RE_TIMESTRING = re.compile(r'\d{14}')  # e.g. 20160512153302
DEFAULT_CHECK_PATH = '/'
DATA_STALE_THRESHOLD = datetime.timedelta(hours=1)

NCDU_JOB_LOCK_PATTREN = NCDU_JOB_LOCK_PREFIX + '-{escped_check_path}'


def check_disk(args):
    data_path_avail_cmd = 'df -BM %s' % args.ncdu_data_path
    res = process.call(data_path_avail_cmd)
    stdout, stderr = res['stdout'], res['stderr']
    if stderr:
        raise RuntimeError(stderr)
    else:
        avail = int(stdout.split()[10].strip('M'))
        if avail < 50:
            raise RuntimeError("Not enough space(<50M) for storing datafile on %s, "
                               "disk usage scanning stopped." % args.ncdu_data_path)
        else:
            lock_file = NCDU_JOB_LOCK_PATTREN.format(escped_check_path=args.escaped_check_path)
            job_lock = open(lock_file, 'w')
            try:
                fcntl.flock(job_lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                logger.exception("sa-disk check job already in running.")
            else:
                job_lock.close()
                time_now_string = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
                data_file_path = os.path.join(args.ncdu_data_path,
                                              'ncdu-export-%s-%s.gz' % (args.escaped_check_path,
                                                                        time_now_string))
                args.data_file = data_file_path
                with open(data_file_path, 'w'):
                    pass
                check_disk_cmd = ('/usr/bin/flock -n %s ncdu -0xo- %s'
                                  ' | gzip >%s'
                                  % (lock_file, args.check_path, data_file_path))
                logger.debug('Running: %s', check_disk_cmd)
                p = process.call(check_disk_cmd, nonblock=True, shell=True)
                logger.info("Checking %s using ncdu, and exporting datafile %s on background"
                            % (args.check_path, data_file_path))

                if args.block:
                    p.wait()
                    read_data(args)


def read_data(args):
    data_file_path = args.data_file
    read_log_cmd = 'zcat %s | ncdu -f-' % data_file_path  # Security Problem
    if not args.topn:
        logger.info("About to open ncdu ui, read datafile %s" % data_file_path)
        prompt_input("Press any key to continue...")
        process.call(read_log_cmd, shell=True)
    if args.topn:
        logger.info("About to calc top %d huge items in depth %s." % (args.topn, args.max_depth))
        stdout = process.call('zcat %s' % data_file_path)['stdout']
        rawfscontent = json.loads(to_str(stdout, errors='ignore'))[3]
        logger.debug('calc top_huge_dirs_from_ncdu...')
        ret = top_huge_dirs_from_ncdu(args.topn, rawfscontent, max_depth=args.max_depth)
        logger.debug('done top_huge_dirs_from_ncdu')

        print("Top %d huge items in depth %s:" % (args.topn, args.max_depth))
        for k, v in sorted(ret.items(), key=lambda kv: kv[1], reverse=True):
            humanize_size = humanize.naturalsize(v, binary=True)
            print('%s  %s' % (humanize_size, k))


def escape_path(path):
    return os.path.abspath(path).replace('/', '%')


def find_latest_exported_data(args):
    logger.info('Finding lastest ncdu exported data file from %s' % args.ncdu_data_path)
    find_latest_cmd = "ls -lt %s | grep ncdu-export-%s-" % (args.ncdu_data_path,
                                                            args.escaped_check_path)
    res = process.call(find_latest_cmd, shell=True)
    stdout, stderr = res['stdout'], res['stderr']
    if stderr:
        raise RuntimeError(stderr)
    elif not stdout:
        return None
    else:
        latest_file_name = stdout.splitlines()[0].split()[-1]
        data_file = os.path.join(args.ncdu_data_path, latest_file_name)
        logger.info('Found ncdu exported data file %s' % data_file)

        data_file_timestring = RE_TIMESTRING.findall(data_file)[0]
        data_file_timestamp = datetime.datetime.strptime(data_file_timestring, "%Y%m%d%H%M%S")
        if (not args.force_read
                and data_file_timestamp <= datetime.datetime.now() - DATA_STALE_THRESHOLD):
            logger.warning("But data file is out of date.")
            return None
        return data_file


def select_ncdu_data_path(args):
    for p in (NCDU_EXPORT_DATA_PATH, '/data', '/'):
        if os.path.isdir(p):
            args.ncdu_data_path = p
            return


@require_sa
def clean(args):
    """clean disk, including outdated ncdu exported data file"""
    select_ncdu_data_path(args)
    if not args.ncdu_data_path:
        logger.error('select_ncdu_data_path failed: no ncdu_data_path selected')
        return
    cmd = ("find -L %s -maxdepth 1 -type f -name 'ncdu-export-*' -mtime +1 -delete"
           % args.ncdu_data_path)
    if args.debug:
        logger.debug('execute %s', cmd)
    res = process.call(cmd, shell=True)
    stdout, stderr = res['stdout'], res['stderr']
    if stdout:
        print(stdout)
    if stderr:
        logger.error('find command got stderr: %s', stderr)


@require_sa
def usage(args):
    """
    若没有指定 -p,--ncdu-data-path 则自动选择一个 path
    若指定 -f,--force-check 则进行强制进行 check
    若指定 -r,--data-file 则读指定的 ncdu exported data 文件；反之，自动选择最近的 data 文件，若找不到 1 个小时内的文件，则依然会进行 check
    若指定 -R 则不会进行 check
    """
    args.escaped_check_path = escape_path(args.check_path)
    if not args.ncdu_data_path:
        select_ncdu_data_path(args)
    if args.force_check:
        check_disk(args)
    else:
        if not args.data_file:
            args.data_file = find_latest_exported_data(args)

        if not args.data_file:
            if args.force_read:
                logger.warning("No data file found.")
            else:
                logger.warning("Recent data file not found. Checking disk usage.")
                check_disk(args)
        else:
            read_data(args)


def main():
    """
    e.g.
    sa-disk usage
    sa-disk usage -n 5 -d 3
    sa-disk usage -r /data1/ncdu-export-%-20160513142844.gz
    sa-disk usage -c /tmp
    sa-disk usage -p /data
    sa-disk usage --force-check

    sa-disk clean
    """
    parser = argparse.ArgumentParser(epilog=main.__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-u', '--user', help='LDAP username, use your OS user name by default.')
    parser.add_argument('--debug', action='store_true', help='debug.')

    # compatible with py2 & 3
    subparsers = parser.add_subparsers(help='Sub commands', dest='subparser')
    subparsers.required = True

    usage_parser = subparsers.add_parser('usage', help='Disk usage operation.')
    usage_parser.add_argument('-p', '--ncdu-data-path', help='Path to store ncdu exported data',
                              default=None)
    usage_parser.add_argument('-r', '--data-file', metavar='DATA_FILE',
                              help='Read ncdu export data file.')
    usage_parser.add_argument('-n', '--topn', type=int, nargs='?',
                              help='Specify top N huge files for output.', const=5)
    usage_parser.add_argument('-d', '--max-depth', type=int,
                              help='Max path depth to calc top N.', default=3)
    usage_parser.add_argument('-c', '--check-path', metavar='CHECK_PATH',
                              help='Check disk and export ncdu data file.',
                              default=DEFAULT_CHECK_PATH)
    usage_parser.add_argument('-f', '--force-check', action='store_true',
                              help='force check, do not use existing exported file')
    usage_parser.add_argument('-R', '--force-read', action='store_true',
                              help='force read latest exported data file, do not check')
    usage_parser.add_argument('-b', '--block', action='store_true',
                              help='disk usage check block mode.')
    usage_parser.set_defaults(func=usage)

    clean_parser = subparsers.add_parser('clean', help='Clean disk, including outdated ncdu '
                                                       'exported data file.')
    clean_parser.set_defaults(func=clean)

    args = parser.parse_args()
    args.user = args.user or get_os_username()

    log_level = logging.DEBUG if args.debug else logging.INFO
    # logger.setLevel(log_level)
    # ch = logging.StreamHandler()
    # ch.setLevel(log_level)
    # formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
    # ch.setFormatter(formatter)
    # logger.addHandler(ch)

    logging.basicConfig(level=log_level,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    args.func(args)



# coding: utf-8

from __future__ import absolute_import

import logging
import smtplib
from email.mime.text import MIMEText

from sa_tools_core.utils import get_config
from sa_tools_core.consts import (SMTP_SERVER, SMTP_SERVER_PORT, SMTP_CREDENTIALS_CONFIG,
                                  SMTP_SSL, SYSADMIN_EMAIL)

logger = logging.getLogger(__name__)


def send_mail(to_addrs, content, subject='Sent via sa-tools', from_addr=SYSADMIN_EMAIL):
    server_info = (SMTP_SERVER, SMTP_SERVER_PORT)
    smtp = smtplib.SMTP_SSL(*server_info) if SMTP_SSL else smtplib.SMTP(*server_info)
    if SMTP_CREDENTIALS_CONFIG:
        user, password = get_config(SMTP_CREDENTIALS_CONFIG).split(':')
        smtp.login(user, password)
    msg = MIMEText(content, 'plain')
    msg['Subject'] = subject
    msg['From'] = from_addr
    msg['To'] = ', '.join(to_addrs)

    try:
        smtp.sendmail(from_addr, to_addrs, msg.as_string())
        logger.info('sent email to %s, title: %s', to_addrs, subject)
    except Exception:
        raise
    finally:
        smtp.quit()



# coding: utf-8

from io import open
from setuptools import setup, find_packages

version = '0.5.0'

requirements = [
    'setuptools',
    'six>=1.12.0',
    'sentry-sdk',

    # sa-notify
    'wechatpy',
    'pylark',
    'requests',

    # sa-dns
    'pydnspod2',

    # sa-access
    'elasticsearch>=7.0.0',

    # sa-disk
    'humanize',
]

# see also https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies  # NOQA
extras_require = {
    # sa-script
    'script': [
        'ansible>=2.8',
        # 'paramiko',  # sa-script use ssh by default
        'progressbar2',
        'terminaltables',
    ],
    # sa-icinga
    'icinga': [
        'Mako',
        'icinga2py',
        'inflect',
    ],
    # sa-tc
    'tencentcloud': [
        'tencentcloud-sdk-python',
        'PyYAML',
        'inflection',
    ],
}

entry_points = """
    [console_scripts]
    sa-uptime = sa_tools_core.uptime:main
    sa-dns = sa_tools_core.dns:main
    sa-disk = sa_tools_core.disk:main
    sa-notify = sa_tools_core.notify:main
    sa-access = sa_tools_core.access:main
    sa-script = sa_tools_core.script:main [script]
    sa-icinga = sa_tools_core.icinga:main [icinga]
    sa-tc = sa_tools_core.tc:main [tencentcloud]
    sa-github = sa_tools_core.github:main
    # sa-node = sa_tools_core.node:main
    # sa-lvs = sa_tools_core.lvs:main
    # sa-rsync = sa_tools_core.rsync:main
"""

scripts = []

setup(
    name='sa-tools-core',
    version=version,
    description="SA Tools Core",
    long_description=open("README.md", encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    # Get more strings from
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='https://github.com/douban/sa-tools-core',
    keywords=['sa-tools', 'sysadmin', 'command line tools'],
    author='sysadmin',
    author_email='sysadmin@douban.com',
    license='BSD',
    packages=find_packages(exclude=['examples*', 'tests*']),
    include_package_data=True,
    zip_safe=False,
    entry_points=entry_points,
    scripts=scripts,
    install_requires=requirements,
    extras_require=extras_require,
)  # NOQA



# coding: utf-8

from __future__ import print_function

import os
import getpass
import sys
import six
import socket
import binascii
import importlib
from functools import wraps

from sa_tools_core.consts import CONFIG_DIR

HOSTS_FILE = '/etc/hosts'
HOSTS_WAN_FILE = '/etc/hosts.wan'

ip_hostname_cache = {}


def get_os_username():
    return getpass.getuser()


def get_config(config_name):
    try:
        with open(os.path.join(CONFIG_DIR, config_name)) as f:
            config = f.read()
    except IOError as e:
        if e.errno == 2:
            print('Error: get config failed, you shall not execute this program on this machine!', file=sys.stderr)
            sys.exit(1)
        raise
    return config.strip()


def resolve_hostname(hostname):
    return socket.gethostbyname(hostname)


def resolve_ip(ip):
    from sh import grep

    if ip_hostname_cache.get(ip):
        return ip_hostname_cache[ip]
    hostname = ip
    # tricky
    if ip.startswith('192.') or ip.startswith('10.'):
        hosts_files = (HOSTS_FILE,)
    else:
        hosts_files = (HOSTS_WAN_FILE,)
    for hosts_file in hosts_files:
        try:
            out = grep('^[^#]*%s[^0-9]' % ip, hosts_file, E=True)
            if out:
                hostname = out.split()[-1]
                break
        except Exception:
            pass
    ip_hostname_cache[ip] = hostname
    return hostname


def reverse_func(f):
    """ reverse function Boolean result """
    @wraps(f)
    def wrapper(*args, **kwargs):
        return not f(*args, **kwargs)

    return wrapper


def ipv6_addr_to_tinydns_generic(ipv6_addr):
    rdata = ''
    hexlify_ipv6_addr = binascii.hexlify(socket.inet_pton(socket.AF_INET6, ipv6_addr))
    split_hexlify_ipv6_addr = [hexlify_ipv6_addr[i:i + 2] for i in range(0, len(hexlify_ipv6_addr), 2)]
    for part in split_hexlify_ipv6_addr:
        rdata += '\\%03o' % int(part, 16)
    return rdata


def to_unicode(s, encoding='utf-8', errors='strict'):
    return six.ensure_text(s, encoding=encoding, errors=errors)


def to_str(s, encoding='utf-8', errors='strict'):
    '''
    in python2, convert s to str if s is unicode
    in python3, convert s to str if s is bytes
    '''
    return six.ensure_str(s, encoding=encoding, errors=errors)


def output(*args, **kw):
    '''
    try to output as UTF-8 encoding strings, to a terminal or a file
    '''
    print(*[to_str(a) for a in args], **kw)


def prompt_input(prompt=None):
    six.moves.input(prompt)


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)

    def __getattr__(self, attr):
        return self.get(attr, self.get('_default_value'))

    def __setattr__(self, attr, value):
        self[attr] = value


def props(cls):
    return [i for i in cls.__dict__ if not i.startswith('_')]


def i2ip(i):
    i = int(i)
    return '%s.%s.%s.%s' % ((i >> 24) % 256, (i >> 16) % 256, (i >> 8) % 256, i % 256)


def import_string(s):
    mod, attr = s.split(':')
    mod = importlib.import_module(mod)
    return getattr(mod, attr) if attr else mod



# coding: utf-8

import logging
from argparse import ArgumentParser

from sa_tools_core.libs.permission import require_sa
from sa_tools_core.libs.process import process
from sa_tools_core.utils import get_os_username


logger = logging.getLogger(__name__)


@require_sa
def uptime(args):
    logger.info('getting uptime info.')
    print(process.uptime()['stdout'])


def main(args=None):
    parser = ArgumentParser()
    parser.add_argument('-u', '--user', help='LDAP username')

    args = parser.parse_args(args)

    args.user = args.user or get_os_username()

    logging.basicConfig(level=logging.WARNING,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')

    uptime(args)
